#include "stdinc.h"
#include <assert.h>
#include "thread.h"
#include "handle.h"
#include "attr.h"
#include "smalloc.h"
#include "fragment.h"

#ifndef WINDOWS
#include <pwd.h>
#include "geckonfs.h"
#endif

#define ATTR_FLAG_CONSIST 1

typedef struct TAttributes *pAttributes;
struct TAttributes {
  struct fattr nfsattr;
  int Flags;
};

MUTEX_TYPE MutexAttr;
Hash_Table *pHashAttr;

void strlower(char *x)
{
  while (*x) {
    *x=tolower(*x);
    x++;
  }
}

void AttrInit(void)
{
  CREATE_MUTEX(MutexAttr);

  pHashAttr = (Hash_Table*) malloc(sizeof(Hash_Table));

  Hash_InitTable(pHashAttr,0,FHSIZE/4);
}

pAttributes AttrAlloc(void)
{
  pAttributes attr;

  attr = (pAttributes) SafeMalloc(sizeof(struct TAttributes));

  memset(attr,0,sizeof(struct TAttributes));

  return attr;
}

void AttrLock(void)
{
  MUTEX_LOCK(MutexAttr);
}

void AttrUnlock(void)
{
  MUTEX_UNLOCK(MutexAttr);
}

void AttrGenerate(char *handle, pAttributes attr)
{
  int ftype = HandleGetType(handle);
  char *url;
  char *metaptr;
  int inode=0;

  if (HandleToUrl(handle,FHSIZE,&url,&inode,&metaptr))
    SafeFree(url);

  attr->nfsattr.fileid = inode;
  attr->nfsattr.atime.seconds = 1;
  attr->nfsattr.mtime.seconds = 1;
  attr->nfsattr.ctime.seconds = 1;
  attr->nfsattr.size = 4096;
  attr->nfsattr.blocksize = 4096;
  attr->nfsattr.blocks = 1;
  attr->nfsattr.rdev = 0;
  attr->nfsattr.fsid = 0;
#ifndef WINDOWS
  attr->nfsattr.uid = NobodyPasswordEntry.pw_uid;
  attr->nfsattr.gid = NobodyPasswordEntry.pw_gid;
#endif
  attr->nfsattr.nlink=2;

  attr->Flags = 0;

  switch (ftype) {
    case FLAG_ROOT:
    case FLAG_HEADERSET:
    case FLAG_HARDLINK:
      attr->nfsattr.type=NFDIR;
      attr->nfsattr.mode=040777L;
      attr->Flags|=ATTR_FLAG_CONSIST;
      break;

    case FLAG_SYMLINK:
      attr->nfsattr.type=NFLNK;
      attr->nfsattr.mode=0120777L;
      attr->Flags|=ATTR_FLAG_CONSIST;
      break;

    case FLAG_INFO:
    case FLAG_GENERATOR:
    case FLAG_CONTENTS:
    case FLAG_HEADERS:
      attr->nfsattr.type=NFREG;
      attr->nfsattr.mode=0100666L;
      break;
  }
}

int AttrHashFind(char *handle, int cancreate, Hash_Entry **h, pAttributes *attr, int *isnew)
{
  AttrLock();

  (*isnew) = FALSE;

  (*h) =  cancreate ? Hash_CreateEntry(pHashAttr,handle,isnew) : Hash_FindEntry(pHashAttr,handle);

  if (*h==NULL) {
    AttrUnlock();
    return FALSE;
  }

  if (*isnew) {
    (*attr) = AttrAlloc();

    AttrGenerate(handle,(*attr));

    Hash_SetValue(*h,(*attr));
  } else
    (*attr) = (pAttributes) Hash_GetValue(*h);

  return TRUE;
}


void AttrLookup(char *handle, struct fattr *attr)
{
  int isnew = FALSE;
  Hash_Entry *h;
  pAttributes attr2;

  assert(AttrHashFind(handle,TRUE,&h,&attr2,&isnew));

  *attr = attr2->nfsattr;

  AttrUnlock();
}

void AttrIncrementTime(char *handle)
{
  Hash_Entry *h;
  pAttributes attr;
  int isnew;

  if (AttrHashFind(handle,FALSE,&h,&attr,&isnew)) {
    attr->nfsattr.atime.seconds++;
    attr->nfsattr.mtime.seconds++;
    attr->nfsattr.ctime.seconds++;
    AttrUnlock();
  }
}

void AttrMakeConsist(char *handle, struct fattr *dest)
{
  pHttpData data;
  Hash_Entry *h;
  pAttributes attr;
  int isnew;
  int length=-1;

  data=HttpDataAlloc(HTTPDATA_DEFAULT_SIZE);

  if (HandleGetType(handle)==FLAG_GENERATOR) {
    length=0;
  } else if (HandleGetType(handle)==FLAG_INFO) {
    HandleGetInfo(handle,FHSIZE,data->Bytes,&data->Count);
    data->ResultPtr=data->Bytes;
    length=data->Count;
  } else if ((GetHeaders(handle,0,data->Size,data,0)) && (data->Count>128)) {
    char *clptr;

    *(data->ResultPtr+data->Count)='\0';
    strlower(data->ResultPtr);

    if (HandleGetType(handle)==FLAG_HEADERS) {
      length = data->Count-128;  // minus 128 to put boundary in middle of pad
    } else {
      clptr = strstr(data->ResultPtr,"content-length:");
      if (clptr!=NULL) {
        clptr+=15;
        while (isspace(*clptr))
          clptr++;
        length = atoi(clptr);
      }
    }
  }

  // if we couldn't get a length, and this is a contents file, then
  // try reading the whole thing to get the length.
  if ((length<0) && (HandleGetType(handle)==FLAG_CONTENTS)) {
    length=0;
    while ((GetBytes(handle,length,data->Size,data,0)) && (data->Count>0))
      length+=data->Count;
  }

  // if length<0, then we have a problem (squid is down?) so set length=0
  // because its the best we can do.
  if (length<0)
    length=0;

  if (AttrHashFind(handle,FALSE,&h,&attr,&isnew)) {
    attr->nfsattr.size = length;
    attr->nfsattr.blocks = (length/attr->nfsattr.blocksize);
    attr->Flags |= ATTR_FLAG_CONSIST;

    if (length%attr->nfsattr.blocksize!=0)
      attr->nfsattr.blocks++;

    if (dest!=NULL)
      *dest=attr->nfsattr;

    AttrUnlock();
  }

  HttpDataFree(data);
}

int AttrLookupConsistant(char *handle, struct fattr *nfsattr)
{
  Hash_Entry *h;
  pAttributes attr;
  int isnew;

  if (AttrHashFind(handle,TRUE,&h,&attr,&isnew)) {
    int needconsist;

    *nfsattr = attr->nfsattr;
    needconsist = (attr->Flags&ATTR_FLAG_CONSIST)==0;
    AttrUnlock();

    if (needconsist)
      AttrMakeConsist(handle,nfsattr);

    return TRUE;
  } else
    return FALSE;
}

void AttrSetPermissions(char *handle, struct sattr *nfsattr)
{
   Hash_Entry *h;
   pAttributes attr;
   int isnew;

   if (AttrHashFind(handle,TRUE,&h,&attr,&isnew)) {
     if (nfsattr->uid!=0xFFFFFFFF)
       attr->nfsattr.uid = nfsattr->uid;

     if (nfsattr->mode!=0xFFFFFFFF)
       attr->nfsattr.mode = nfsattr->mode;

     if (nfsattr->gid!=0xFFFFFFFF)
       attr->nfsattr.gid = nfsattr->gid;

     AttrUnlock();
   }
}

