#include "stdinc.h"
#include <sys/times.h>
#include <errno.h>
#include "html.h"
#include "buffer.h"
#include "httpcon.h"

#define CONTENTS_NAME ".contents"

char MountPoint[256];
char HttpHost[256];
int HttpPort;
int UseNfs;
int Iterations;
int FileSize;
int BlockSize;
int Diagnostics;
int AllBench;
int UniqueFiles;
int OpenFirst;
int TimeoutTest;
int UseLocalFiles;

int IterationNumber,FileNumber,UniqueIterationNumber=0;

char *DataBuffer;

void BenchFileHttp(void)
{
  SOCKET sock = InitHttpSocket(HttpHost,HttpPort);
  char msg[256];
  int bytesrecv, totalrecv = 0;

  if (sock==INVALID_SOCKET) {
    printf("failed to initialize http socket!\n");
    return;
  }

  sprintf(msg,"GET http://pe39.cs.arizona.edu/test%d HTTP/1.0\r\n\r\n",FileSize);
  send(sock,msg,strlen(msg),0);
//  shutdown(sock,1); // shutdown transmitter

  while ((bytesrecv=recv(sock,DataBuffer,BlockSize,0))>0) {
    if (Diagnostics)
      printf("receive %d\n",bytesrecv);
    totalrecv +=bytesrecv;
  }

  if (bytesrecv<0) {
    perror("receive error");
  }

/*  if (totalrecv>FileSize)
    printf("too many bytes? %d/%d!\n",totalrecv,FileSize);
  else */if (totalrecv<FileSize)
    printf("not enough bytes %d/%d?!\n",totalrecv,FileSize);

  closesocket(sock);
}

int OpenFileNfs(void)
{
  char msg[256];
  char path[256];
  int filedes;

  if (UseLocalFiles)
    strcpy(path,"local_file/testcase");
  else
    strcpy(path,"pe39.cs.arizona.edu");

  if (FileNumber==-1)
    sprintf(msg,"%s/%s/test%d/%s",MountPoint,path,FileSize,CONTENTS_NAME);
  else
    sprintf(msg,"%s/%s/test%d/%s#%d",MountPoint,path,FileSize,CONTENTS_NAME,FileNumber);

//  fprintf(stdout,"%s\n",msg);

  filedes = open(msg,O_RDONLY,0);
  if (filedes<0) {
    printf("failed to open NFS file!\n");
    return;
  }

  return filedes;
}

void BenchFileNfs(int filedes)
{
  int bytesread, totalread=0;
  int timeout = 1;

  if (filedes<0)
    return;

  while ((bytesread=read(filedes,DataBuffer,BlockSize))==BlockSize) {
    if (bytesread<0) 
      printf("read error: bread=%d errno=%d\n",bytesread,errno);

    if (TimeoutTest) {
      printf("sleep(%d)\n",timeout);
      sleep(timeout);
      timeout*=2;
    }

    totalread+=bytesread;
  }
//  fprintf(stderr,"totalread=%d\n",totalread);

  close(filedes);
}

long DoOneIteration(void)
{
  struct timeval start, finish;
  struct timezone tz;

  gettimeofday(&start,&tz);

  if (UseNfs) {
    if (OpenFirst) {
      int filedes = OpenFileNfs();

      gettimeofday(&start,&tz);
      BenchFileNfs(filedes);
      gettimeofday(&finish,&tz);
    } else
      gettimeofday(&start,&tz);
      BenchFileNfs(OpenFileNfs());
      gettimeofday(&finish,&tz);
    }
  else {
    gettimeofday(&start,&tz);
    BenchFileHttp();
    gettimeofday(&finish,&tz);
  }

  return (finish.tv_sec-start.tv_sec)*1000000L+(finish.tv_usec-start.tv_usec);
}

void RunBench(void)
{
  struct timeval start, finish;
  struct timezone tz;
  long times[256]; /* up to 256 iterations */
  long count, sum;
  long average1, average2;
  long deviation;
  int i;
  int throwaway;

  if (Iterations>256) {
    printf("too many iterations\n");
    return;
  }

  FileNumber = 1000; // make sure our 'warm-up' iteration doesn't use a real IterationNumber

  DoOneIteration();

  IterationNumber = 0;

  gettimeofday(&start,&tz);
  while (IterationNumber<Iterations) {
    FileNumber = (UniqueFiles ? UniqueIterationNumber : -1);

    times[IterationNumber] = DoOneIteration();

    IterationNumber++;
    UniqueIterationNumber++;
  }
  gettimeofday(&finish,&tz);

  /* compute the original average */
  sum = 0;
  for (i=0; i<Iterations; i++)
    sum+=times[i];
  average1 = sum/Iterations;

  /* compute the average again, throwing out bad values */
  count=0;
  sum=0;
  throwaway=0;
  for (i=0; i<Iterations; i++) {
    deviation = average1-times[i];

    if (deviation<0)
      deviation*=-1;

    if (FALSE && (deviation>0.25*average1)) {
      printf("%d: %d thrown out\n",i,times[i]);
      throwaway++;
    } else {
      count++;
      sum+=times[i];
    }
  }
  if (throwaway>Iterations/4) {
    average2 = 0;
  } else
    average2 = (count==0 ? 0 : sum/count);

//  average2=((finish.tv_sec-start.tv_sec)*1000000L+(finish.tv_usec-start.tv_usec))/Iterations;
  printf("%s%s%s %d %0.1f\n",OpenFirst ? "O" : "",UniqueFiles ? "" : "C",UseNfs ? "NFS" : "HTTP",FileSize,average2/1000.0);
}

void DoTod(void)
{
  struct timeval cur, last;
  struct timezone tz;

  while (TRUE) {
    gettimeofday(&cur,&tz);
    if ((cur.tv_sec!=last.tv_sec) || (cur.tv_usec!=last.tv_usec)) {
      printf("%ul %ul\n",cur.tv_sec,cur.tv_usec);
      last = cur;
    }
  }
}

static struct { char *name, *use ; } Args[] = {
  {"-h",        "Print usage information"},
  {"-m",        "Specify mount point -m xxx (default is /mnt/web)"},
  {"-s",        "Specify http server -s xxx (default is 198.0.0.100)"},
  {"-p",        "Specify http port number -p xxx (default is 80)"},
  {"-n",        "Use NFS protocol (default is http)"},
  {"-i",	"Iterations per cycle -i xxx (default is 1)"},
  {"-f",	"File size in bytes -f xxx (default is 1024)"},
  {"-d",	"Include diagnostic messages"},
  {"-k",	"Set read() block size"},
  {"-a",	"Run all benchmarks (512b, 1024b, 2048b, ..., 256k"},
  {"-tod",	"Show timeofday calculation"},
  {"-c",	"Do not use unique files (test caching)"},
  {"-timeout",	"NFS Handle timeout test"},
  {"-l",	"Use local files"},
  {NULL, NULL}
  } ;

static void ProcessCommandLine( int argc, char *argv[] ) {
  int a, f ;
  char buf[256];

  strcpy(MountPoint,"/tmp/web");
  strcpy(HttpHost,"gecko.cs.arizona.edu");
  HttpPort = 3128;
  UseNfs = FALSE;
  Iterations = 1;
  FileSize = 1024;
  BlockSize = 32768;
  Diagnostics = FALSE;
  AllBench = FALSE;
  UniqueFiles = TRUE;
  OpenFirst = FALSE;
  TimeoutTest = FALSE;
  UseLocalFiles = FALSE;

  for( a = 1 ; a < argc ; a++ ) {
    char *cmdname = argv[a];

    for( f = 0 ; Args[f].name != NULL ; f++ )
      if( stricmp(Args[f].name, cmdname) == 0 ) break ;

    switch( f ) {
      case 0: break; /* usage information */

      case 1: if (*argv[++a]!='\0')
                strcpy(MountPoint,argv[a]);
	      break;

      case 2: if (*argv[++a]!='\0')
      		strcpy(HttpHost,argv[a]);
	      break;

      case 3: if (*argv[++a]!='\0')
                HttpPort = atoi(argv[a]);
	      break;

      case 4: UseNfs = TRUE;
	      break;

      case 5: if (*argv[++a]!='\0')
                Iterations = atoi(argv[a]);
	      break;

      case 6: if (*argv[++a]!='\0')
                FileSize = atoi(argv[a]);
	      break;

      case 7: Diagnostics=TRUE;
      	      break;

      case 8: if (*argv[++a]!='\0')
                BlockSize = atoi(argv[a]);
      	      break;

      case 9: AllBench = TRUE;
      	      break;

      case 10: DoTod();
      	       break;

      case 11: UniqueFiles = FALSE;
      	       break;

      case 12: TimeoutTest = TRUE;
      	       break;

      case 13: UseLocalFiles = TRUE;
      	       break;
    }
  }
}

void RunAllBench()
{
  int j;
  long maxsize = 1024L*1024L;

  UniqueFiles = TRUE;
  OpenFirst = FALSE;

  // http
  FileSize = 256;
  UseNfs = FALSE;
  while (FileSize<=maxsize) {
    RunBench();
    FileSize*=2;
  }

  // nfs
  FileSize = 256;
  UseNfs = TRUE;
  while (FileSize<=maxsize) {
    RunBench();
    FileSize*=2;
  }

  // onfs
  OpenFirst = TRUE;
  FileSize = 256;
  UseNfs = TRUE;
  while (FileSize<=maxsize) {
    RunBench();
    FileSize*=2;
  }

  // cnfs
  OpenFirst = FALSE;
  UniqueFiles=FALSE;
  FileSize = 256;
  UseNfs = TRUE;
  while (FileSize<=maxsize) {
    RunBench();
    FileSize*=2;
  }

}

int main(int argc, char **argv)
{
  ProcessCommandLine(argc,argv);

  DataBuffer = (char*) malloc(BlockSize+1); /* +1 just in case */

  if (AllBench)
    RunAllBench();
  else
    RunBench();

  return 0;
}
