#include "stdinc.h"
#include <assert.h>
#include "thread.h"
#include "handle.h"
#include "dircache.h"
#include "url.h"
#include "smalloc.h"
#include "buffer.h"
#include "parse.h"
#include "inode.h"

Hash_Table *pHashDir;
MUTEX_TYPE MutexDir;
pChildNode ChildNodeFreeList = NULL;
pDirNode DirNodeFreeList = NULL;

void DirInit()
{
  pHashDir = (Hash_Table*) malloc(sizeof(Hash_Table));

  Hash_InitTable(pHashDir,0,FHSIZE/4);

  CREATE_MUTEX(MutexDir);
}

void DirLock(void)
{
  MUTEX_LOCK(MutexDir);
}

void DirUnlock(void)
{
  MUTEX_UNLOCK(MutexDir);
}

pChildNode ChildNodeAlloc(void)
{
  pChildNode result;

  if (ChildNodeFreeList==NULL)
    result = (pChildNode) SafeMalloc(sizeof(struct TChildNode));
  else {
    result = ChildNodeFreeList;
    ChildNodeFreeList=ChildNodeFreeList->Next;
  }

  memset(result,0,sizeof(struct TChildNode));
  return result;
}

void ChildNodeFree(pChildNode node)
{
  node->Next = ChildNodeFreeList;
  ChildNodeFreeList=node;
}

pDirNode DirNodeAlloc(void)
{
  pDirNode result;

  if (DirNodeFreeList==NULL)
    result = (pDirNode) SafeMalloc(sizeof(struct TDirNode));
  else {
    result = DirNodeFreeList;
    DirNodeFreeList=DirNodeFreeList->Next;
  }

  memset(result,0,sizeof(struct TDirNode));
  return result;
}

int FindChild(pChildNode first, char *name, fhandle dest)
{
  while (first!=NULL) {
    if (strcmp(first->Name,name)==0) {
      HandleCopy(dest,first->Handle);
      return TRUE;
    }
    first=first->Next;
  }

  return FALSE;
}

pChildNode DirAddChildToList(pChildNode *list, char *name, fhandle childhandle)
{
  pChildNode cur = *list;

  while ((cur!=NULL) && (strcmp(cur->Name,name)!=0))
    cur=cur->Next;

  if (cur==NULL) {
    cur = ChildNodeAlloc();
    cur->Next=*list;
    *list=cur;
  }

  cur->Name = SafeStrDup(name);
  HandleCopy(cur->Handle,childhandle);

  return cur;
}

int DirHashFind(fhandle parent, int cancreate, Hash_Entry **h, pDirNode *dir, int *isnew)
{
  (*isnew)=FALSE;

  DirLock();

  (*h) = (cancreate) ? Hash_CreateEntry(pHashDir,parent,isnew) : Hash_FindEntry(pHashDir,parent);

  if (*h==NULL) {
    DirUnlock();
    return FALSE;
  }

  if (*isnew) {
    (*dir) = DirNodeAlloc();

    (*dir)->HardLinks=NULL;
    (*dir)->SymLinks=NULL;
    (*dir)->HashEntry=(*h);
    (*dir)->Parent=NULL;
    (*dir)->Self=DirAddChildToList(&((*dir)->SymLinks),".",parent);

    Hash_SetValue(*h,*dir);
  } else
    (*dir) = (pDirNode) Hash_GetValue(*h);

  return TRUE;
}

void DirAddChild(fhandle parent, fhandle childhandle, char *name, int symbolic)
{
  int isnew=FALSE;
  Hash_Entry *h;
  pDirNode dir;

  assert(DirHashFind(parent,TRUE,&h,&dir,&isnew));

  DirAddChildToList(symbolic ? &(dir->SymLinks) : &(dir->HardLinks),name,childhandle);

  DirUnlock();
}

int DirLookup(fhandle parent, char *name, fhandle dest)
{
  Hash_Entry *h;
  pDirNode dir;
  int result;
  int isnew=FALSE;

  if (DirHashFind(parent,FALSE,&h,&dir,&isnew)) {
    result =  FindChild(dir->HardLinks,name,dest) || FindChild(dir->SymLinks,name,dest);
    DirUnlock();
  } else
    result = FALSE;

  return result;
}

int AddEntry(pentry *first, pentry *last, pChildNode child, int index)
{
  pentry entry;
  char *url;
  int inode;
  char *metaptr;

  if (!HandleToUrl(child->Handle,FHSIZE,&url,&inode,&metaptr))
    return FALSE;

  entry = (pentry) malloc(sizeof(struct entry));
  assert(entry!=NULL);

  entry->fileid = inode;
  entry->name = SafeStrDup(child->Name);
  entry->nextentry = NULL;
  memcpy(entry->cookie,&index,4);

  if (*last!=NULL)
    (*last)->nextentry = entry;
  else
    (*first)=entry;

  *last = entry;

  SafeFree(url);

  return TRUE;
}

int DirRead(fhandle parent, int start, pentry *first)
{
  pDirNode dir;
  pChildNode child;
  int index;
  int outcount;
  pentry last=NULL;
  int isnew = FALSE;
  Hash_Entry *h;

  *first=NULL;

  if (DirHashFind(parent,FALSE,&h,&dir,&isnew)) {
    index = 0;
    outcount = 0;

    for (child=dir->HardLinks; child!=NULL; child=child->Next) {

/*      int inode;
      char *url,*metaptr;
      HandleToUrl(child->Handle,FHSIZE,&url,&inode,&metaptr);
      ShowHandle(stderr,child->Handle);
      fprintf(stderr,"%d %d %d %s %s\n",index,start,inode,child->Name,url);
      SafeFree(url);*/

      if ((index>=start) && (AddEntry(first,&last,child,index+1)))
        outcount++;
      index++;
    }

    for (child=dir->SymLinks; child!=NULL; child=child->Next) {

/*      int inode;
      char *url,*metaptr;
      HandleToUrl(child->Handle,FHSIZE,&url,&inode,&metaptr);
      ShowHandle(stderr,child->Handle);
      fprintf(stderr,"%d %d %d %d %s %s\n",index,start,inode,HandleGetType(child->Handle),child->Name,url);
      SafeFree(url);*/

      if ((index>=start) && (AddEntry(first,&last,child,index+1)))
        outcount++;
      index++;
    }
    DirUnlock();
  } else
    outcount = -1;

  return outcount;
}

void DirSetParent(fhandle hdir, fhandle hparent, char *name)
{
  pDirNode dir;
  int isnew = FALSE;
  int type = HandleGetType(hdir);
  Hash_Entry *h;

  if ((type!=FLAG_HARDLINK) && (type!=FLAG_HEADERSET))
    return;

  if (strcmp(name,"..")==0)
    return;

  if (DirHashFind(hdir,TRUE,&h,&dir,&isnew)) {
    if (dir->Parent==NULL)
      dir->Parent = DirAddChildToList(&(dir->HardLinks),"..",hparent);
    else
      HandleCopy(dir->Parent->Handle,hparent);

    DirUnlock();
  }
}

int DirAddContentsAndHeaders(pDirNode dir, char *url, int parentinode, char *headers)
{
  char *metaptr;

/*  if ((dir->Flags&DIR_FLAG_HASDOT)==0) {
    InheritChild(&(dir->SymLinks),".",url,".");
    dir->Flags|=DIR_FLAG_HASDOT;
  }

  if (dir->Flags&DIR_FLAG_HASDOTDOT)==0) {
    InheritChild(&(dir->SymLinks),"..",url.".."); */

  if ((dir->Flags&DIR_FLAG_HASCONTENTS)==0) {
    char *contentsname = ".contents";
    fhandle childhandle;

    UrlToHandle(url,InodeCreate(parentinode,FALSE),childhandle,FHSIZE,&metaptr);
    HeadersToHandle(headers,childhandle,FHSIZE,metaptr);
    HandleInitFileType(childhandle,contentsname);

    dir->Flags|=DIR_FLAG_HASCONTENTS;
    DirAddChildToList(&(dir->SymLinks),contentsname,childhandle);
  }

  if ((dir->Flags&DIR_FLAG_HASHEADERS)==0) {
    char *headersname=".headers";
    fhandle headershandle;

    UrlToHandle(url,InodeCreate(parentinode,FALSE),headershandle,FHSIZE,&metaptr);
    HeadersToHandle(headers,headershandle,FHSIZE,metaptr);
    HandleInitFileType(headershandle,headersname);

    ShowHandle(stderr,headershandle);

    dir->Flags|=DIR_FLAG_HASHEADERS;
    DirAddChildToList(&(dir->SymLinks),headersname,headershandle);
  }

  if ((dir->Flags&DIR_FLAG_HASINFO)==0) {
    char *infoname=".info";
    fhandle infohandle;

    UrlToHandle(url,InodeCreate(parentinode,FALSE),infohandle,FHSIZE,&metaptr);
    HeadersToHandle(headers,infohandle,FHSIZE,metaptr);
    HandleInitFileType(infohandle,infoname);

    ShowHandle(stderr,infohandle);

    dir->Flags|=DIR_FLAG_HASINFO;
    DirAddChildToList(&(dir->SymLinks),infoname,infohandle);
  }

  if ((dir->Flags&DIR_FLAG_HASDOTROOT)==0) {
    CreateDotRoot(&(dir->SymLinks),url,parentinode);
    dir->Flags|=DIR_FLAG_HASDOTROOT;
  }

  return TRUE;
}

void DirParse(char *handle, char *url, int parentinode)
{
  struct Buffer *buf = BufferCreate();
  struct tnode *rootnode = NULL;
  pChildNode newlinks = NULL;
  pChildNode last,*cur;
  int nlink=0, nimage=0, nframe=0;
  Hash_Entry *h;
  pDirNode dir;
  int isnew;

  fprintf(stderr,"b0\n");

  buf->UseNfsHandle=TRUE;
  HandleCopy(buf->NfsHandle,handle);

  fprintf(stderr,"b1\n");

  HtmlParse(&rootnode,buf);

  fprintf(stderr,"b2\n");

  ExtractLinks(&newlinks,rootnode,&nlink,&nframe,&nimage,url,parentinode);

  HtmlDispose(rootnode);

  assert(DirHashFind(handle,TRUE,&h,&dir,&isnew));

  cur = &dir->SymLinks;
  while (*cur!=NULL) {
    if (((*cur)->Flags&CHILD_FLAG_ISDOTLINK)!=0) {
      pChildNode todelete = (*cur);
      (*cur) = (*cur)->Next;

      ChildNodeFree(todelete);
    } else {
      last = *cur;
      cur = &((*cur)->Next);
    }
  }

  last->Next = newlinks;

  dir->Flags|=DIR_FLAG_CONSIST;

  DirUnlock();
}

int DirMakeConsist(char *handle)
{
  Hash_Entry *h;
  int isnew;
  pDirNode dir;
  int needparse;
  char *url, *metaptr, *headers;
  int inode;

  fprintf(stderr,"dirmakeconsist\n");

  assert(DirHashFind(handle,TRUE,&h,&dir,&isnew));

  if ((dir->Flags & DIR_FLAG_ALLCONSIST)==DIR_FLAG_ALLCONSIST) {
    fprintf(stderr,"consist\n");
    DirUnlock();
    return FALSE;
  }

  if (!HandleToUrl(handle,FHSIZE,&url,&inode,&metaptr)) {
    fprintf(stderr,"hufail\n");
    DirUnlock();
    return FALSE;
  }

  if (!HandleToHeaders(handle,FHSIZE,metaptr,&headers)) {
    fprintf(stderr,"hthfail\n");
    SafeFree(url);
    DirUnlock();
    return FALSE;
  }

  DirAddContentsAndHeaders(dir,url,inode,headers);

  needparse = (dir->Flags&DIR_FLAG_CONSIST)!=0;

  DirUnlock();

  if (!needparse) {
    fprintf(stderr,"a1\n");
    DirParse(handle,url,inode); // make consistant
  }

  SafeFree(url);
  SafeFree(headers);

  return TRUE;
}

// returns -1 if no handle present, o/w a count of the returned handles
/*
int DirRead(fhandle parent, int start, fhandle *dest, int maxlist)
{
  pDirNode dir;
  pChildNode child;
  int index;
  int outindex;
  Hash_Entry *h;

  DirLock();

  h = Hash_FindEntry(pHashDir,parent);

  if (h!=NULL) {
    dir = (pDirNode) Hash_GetValue(h);

    outindex=0;
    index=0;

    for (child=dir->HardLinks; child!=NULL; child=child->Next) {
      if ((index>=start) && (outindex<maxlist))
        HandleCopy(dest[outindex++],child->Handle);
      index++;
    }

    for (child=dir->SymLinks; child!=NULL; child=child->Next) {
      if ((index>=start) && (outindex<maxlist))
        HandleCopy(dest[outindex++],child->Handle);
      index++;
    }

    DirUnlock();

    return outindex;
  }

  DirUnlock();

  return -1;
}
*/







