#include "stdinc.h"
#include <assert.h>
#include "thread.h"
#include "fragment.h"
#include "handle.h"
#include "url.h"
#include "smalloc.h"

#ifdef WINDOWS
#include "d:\bc\projects\proxy\hash\Hash.h"
struct sockaddr_un { int sun_family; char *sun_path; };
#define bzero(ptr,len) memset(ptr,0,len)
#define AF_LOCAL 0
#else
#include <sys/un.h>
#include "hash.h"
#endif

#define WORKER_COUNT 32

typedef struct TCircular *pCircular;
struct TCircular {
  int Size;
  int Count;
  int Offset;
  char *Alloc;
  char *Head, *Tail, *Over;
};

typedef struct THttpBuffer *pHttpBuffer;
struct THttpBuffer {
  fhandle Handle;
  struct TCircular DataBuf;
  struct TCircular HeaderBuf;
  int Completed;
  int RefCount;
  MUTEX_TYPE MutexBuffer;
  COND_TYPE SignalRefCountDec;
  COND_TYPE SignalDataReady;
  COND_TYPE SignalRequestPending;
  pHttpBuffer Next;
  int DoingHeaders;
  int RequestPending;
  int KeepaliveSeconds;
  Hash_Entry *HashEntry;
};

char *SquidName;
int SquidPort;
pHttpBuffer HttpBufFreeList = NULL;
Hash_Table HashHttpBuf;
MUTEX_TYPE MutexHttpBuf;
MUTEX_TYPE MutexSocket;
COND_TYPE SignalBufferAvail;
THREAD_TYPE ThreadID[WORKER_COUNT];

// in the case of a refresh, remove the '#' part from the url
void RemovePound(char *x)
{
  char *pound = strchr(x,'#');

  if (pound!=NULL)
    *pound = '\0';
}

pHttpData HttpDataAlloc(int size)
{
  pHttpData buf = (pHttpData) SafeMalloc(sizeof(struct THttpData));

  assert(buf!=NULL);

  buf->Bytes = (char*) SafeMalloc(size+16);
  buf->Size = size;
  buf->ResultPtr = NULL;
  buf->Count = 0;

  return buf;
}

void HttpDataFree(pHttpData buf)
{
  SafeFree(buf->Bytes);
  SafeFree(buf);
}

void CircularInit(pCircular c, int size)
{
  c->Size=size;
  c->Alloc = (char*) malloc(size);
  c->Head = c->Alloc;
  c->Tail = c->Alloc;
  c->Over = c->Alloc+size;
  c->Count = 0;
  c->Offset = 0;
}

void CircularReset(pCircular c)
{
  c->Head=c->Alloc;
  c->Tail=c->Alloc;
  c->Count=0;
  c->Offset=0;
}

void CircularAdd(pCircular buf, char *in, int count)
{
  int i;
  int easy = MIN((buf->Size-buf->Count),count);

  if (easy==1) {
    *(buf->Tail++)=*(in++);
    buf->Count++;
    count--;
  } else if (easy>1) {
    memcpy(buf->Tail,in,easy);
    buf->Count+=easy;
    count-=easy;
    buf->Tail+=easy;
    in+=easy;

    if (buf->Tail==buf->Over)
      buf->Tail=buf->Alloc;
  }

  for (i=0; i<count; i++) {
    *(buf->Tail++)=*in;
    if (buf->Tail==buf->Over)
      buf->Tail=buf->Alloc;

    buf->Head++;
    if (buf->Head==buf->Over)
      buf->Head=buf->Alloc;

    in++;
    buf->Offset++;
  }
}

int CircularGet(char *dest, int offset, int count, pCircular buf)
{
  int delta = offset-buf->Offset;
  int todo = MAX(0,MIN(buf->Count-delta,count));
  char *pfirst = buf->Head+delta;
  int cfirst = MAX(0,MIN(buf->Over-pfirst,todo));
  int deltasecond = MAX(0,pfirst-buf->Over);
  char *psecond = buf->Alloc+deltasecond;
  int csecond = todo-cfirst;

  if (cfirst>0)
    memcpy(dest,pfirst,cfirst);

  if (csecond>0)
    memcpy(dest+cfirst,psecond,csecond);

  return todo;
}

int LookupAddress(char *address, struct in_addr *addr)
{
  unsigned long ip_address;

  #ifdef HTTPCON_STATS
  pStatInfo st=StartStatistics("lookupaddress");
  #endif

  if ((ip_address = inet_addr (address)) != INET_ADDR_FAIL) {
    addr->s_addr = ip_address;
  } else {
    struct hostent *hostentry = NULL;

    #ifdef WINDOWS
    if ((hostentry = gethostbyname(address)) == NULL)
      return -1;
    #else
      char buf[2048];
      struct hostent he1;
      int err=0;
      int result;

      result = gethostbyname_r(address,&he1,buf,sizeof(buf),&hostentry,&err);

      if ((hostentry==NULL) || (result!=0) || (err!=0)) {
        #ifdef HTTPCON_STATS
        RecordStatistics("lookupaddress_fail",st);
        #endif
        return -1;
      }
    #endif

    addr->s_addr = *(ADDRTYPE *) hostentry->h_addr;
  }

  #ifdef HTTPCON_STATS
  RecordStatistics("lookupaddress",st);
  #endif

  return 0;
}

#define UNIXPATH "/tmp/foo.bar"
SOCKET InitHttpSocket(char *servername, int port)
{
#ifdef USE_UNIX_DOMAIN_SOCKET

  // the following code is used to talk to squid via a unix domain socket. 
  // This does require a custom patch to squid to make this work. Using Unix
  // domain sockets provides a significant increase in performance as 
  // compared to talking to squid via TCP/IP

  SOCKET sock;
  struct sockaddr_un servaddr;

  MUTEX_LOCK(MutexSocket);

  sock = socket(AF_UNIX,SOCK_STREAM,0);

  bzero(&servaddr,sizeof(servaddr));
  servaddr.sun_family = AF_LOCAL;
  strcpy(servaddr.sun_path,UNIXPATH);

  if (connect(sock,(struct sockaddr*) &servaddr, sizeof(servaddr))==SOCKET_ERROR) {
    fprintf(stderr,"unable to connect!\n");
    MUTEX_UNLOCK(MutexSocket);
    return -1;
  }

  MUTEX_UNLOCK(MutexSocket);

  return sock;

#else
  struct sockaddr_in inaddr;
  SOCKET sock;

  MUTEX_LOCK(MutexSocket);

  sock = socket(AF_INET, SOCK_STREAM, 0);

  if (sock==INVALID_SOCKET) {
    printf("%s:%d ",servername,port);
    perror("unable to create socket");
//    HttpConError("unable to create socket",servername,port,strerror());
    MUTEX_UNLOCK(MutexSocket);
    return -1;
  }

  if (LookupAddress(servername,&inaddr.sin_addr)!=0) {
    fprintf(stderr,"%s:%d ",servername,port);
    perror("unable to map address");
//    HttpConError("unable to map address",servername,port,strerror());
    closesocket(sock);
    MUTEX_UNLOCK(MutexSocket);
    return -1;
  }

  inaddr.sin_family = AF_INET;
  inaddr.sin_port = htons(port);

  if ((connect (sock,(struct sockaddr *)&inaddr,sizeof(struct sockaddr_in))==SOCKET_ERROR)) {
    printf("%s:%d ",servername,port);
    perror("unable to connect");
    closesocket(sock);
    MUTEX_UNLOCK(MutexSocket);
    return -1;
  }

  MUTEX_UNLOCK(MutexSocket);

  return sock;
#endif
}


int HttpBufInRange(pHttpBuffer buf, pCircular cbuf, int offset, int count)
{
  return (offset>=cbuf->Offset) && ((offset+count<=cbuf->Offset+cbuf->Count) || (buf->Completed));
}

void HttpBufLock(pHttpBuffer buf)
{
  MUTEX_LOCK(buf->MutexBuffer);
}

void HttpBufUnlock(pHttpBuffer buf)
{
  MUTEX_UNLOCK(buf->MutexBuffer);
}

void HttpBufDec(pHttpBuffer buf)
{
  buf->RefCount--;
  COND_BROADCAST(buf->SignalRefCountDec);
}

void HttpBufFree(pHttpBuffer buf)
{
  MUTEX_LOCK(MutexHttpBuf);

  buf->Next=HttpBufFreeList;
  HttpBufFreeList=buf;

  MUTEX_UNLOCK(MutexHttpBuf);
}

void HttpBufInitNode(pHttpBuffer result, fhandle handle)
{
  int isnew=FALSE;

  CircularReset(&(result->DataBuf));
  CircularReset(&(result->HeaderBuf));
  result->RefCount = 0;
  result->Completed = FALSE;
  HandleCopy(result->Handle,handle);

  result->HashEntry = Hash_CreateEntry(&HashHttpBuf,handle,&isnew);
  Hash_SetValue(result->HashEntry,result);
}

pHttpBuffer HttpBufFind(fhandle handle)
{
  Hash_Entry *h;
  pHttpBuffer result;

//  MUTEX_LOCK(MutexHttpBuf);

  h = Hash_FindEntry(&HashHttpBuf,handle);

  if (h==NULL) {
    if (HttpBufFreeList==NULL) {
//      MUTEX_UNLOCK(MutexHttpBuf);
      return NULL;
    }

    result=HttpBufFreeList;
    HttpBufFreeList=HttpBufFreeList->Next;

    HttpBufLock(result);
    HttpBufInitNode(result,handle);

    result->RequestPending=TRUE;
    COND_BROADCAST(result->SignalRequestPending);
  } else {
    result = (pHttpBuffer) Hash_GetValue(h);
    HttpBufLock(result);
  }

  result->RefCount++;

//  MUTEX_UNLOCK(MutexHttpBuf);

  return result;
}

void HttpBufCopyIn(pHttpBuffer dest, char *inbuf, int count)
{
  char last='\0';

  HttpBufLock(dest);

  while ((dest->DoingHeaders) && (count>0)) {
    if ((*inbuf=='\n') && (last=='\n'))
      dest->DoingHeaders=FALSE;
    else if (*inbuf!='\r') {
      last=*inbuf;
      CircularAdd(&(dest->HeaderBuf),inbuf,1);
    }
    count--;
    inbuf++;
  }

  CircularAdd(&(dest->DataBuf),inbuf,count);

  HttpBufUnlock(dest);
}


int SquidGetPartial(pHttpData dest, fhandle handle, int offset, int count)
{
  SOCKET sock;
  char host[1024];
  char url[1024];
  char msg[1024+4096+256];
  char buf[4097];
  int port;
  char *destptr = dest->Bytes;
  int destleft = count;
  int doing_headers;
  int first_headers;
  int len;
  char *datapos;

  char *fullurl, *headers;
  char *metaptr;
  int inode;

  if (!HandleToUrl(handle,FHSIZE,&fullurl,&inode,&metaptr))
    return FALSE;

  RemovePound(fullurl);

  if (!HandleToHeaders(handle,FHSIZE,metaptr,&headers)) {
    SafeFree(fullurl);
    return FALSE;
  }

  if ((strlen(fullurl)>1024) || (strlen(headers)>4096)) {
    SafeFree(fullurl);
    SafeFree(headers);
    return FALSE;
  }

  if (!ParseName(fullurl,host,&port,url)) {
    SafeFree(fullurl);
    SafeFree(headers);
    return FALSE;
  }

  sock = InitHttpSocket(SquidName,SquidPort); //host,port);

  if (sock==INVALID_SOCKET) {
    SafeFree(fullurl);
    SafeFree(headers);
    return FALSE;
  }

  fprintf(stderr,"partial get %s\n",fullurl);
  sprintf(msg,"GET %s HTTP/1.0\r\nProxy-Authorization: gecko\r\nCache-Control: only-if-cached\r\nRange: bytes=%ld-%ld\r\n%s\r\n",
    fullurl,
    offset,
    offset+count,
    headers);

  send(sock,msg,strlen(msg),0);

  doing_headers=TRUE;
  first_headers=TRUE;
  while ((len=recv(sock,buf,sizeof(buf)-1,0))>0) {
    *(buf+len)='\0';
    if (doing_headers) {
      char *x=strstr(buf,"\r\n\r\n");

      if (first_headers) {
        first_headers=FALSE;
        fprintf(stderr,"%0.40s\n",buf);
        if (strncmp(buf+9,"206",3)!=0) {
          SafeFree(fullurl);
          SafeFree(headers);
          closesocket(sock);
          return FALSE;
        }
      }

      if (x!=NULL) {
        doing_headers=FALSE;
        datapos=x+4;
        len-=(x+4-buf);
      }
    } else datapos=buf;

    if (!doing_headers) {
      int todo = MIN(destleft,len);

      memcpy(destptr,datapos,todo);
      destptr+=todo;
      destleft-=todo;
    }
  }

  dest->ResultPtr=dest->Bytes;
  dest->Count=count-destleft;

  closesocket(sock);

  SafeFree(fullurl);
  SafeFree(headers);

  return TRUE;
}

int SquidGetHead(pHttpData dest, fhandle handle, int offset, int count)
{
  SOCKET sock;
  char host[1024];
  char url[1024];
  char msg[1024+4096+256];
  int port;
  char *destptr = dest->Bytes;
  int remain = dest->Size;
  int len;
  int total=0;

  char *fullurl, *headers;
  char *metaptr;
  int inode;

  if (!HandleToUrl(handle,FHSIZE,&fullurl,&inode,&metaptr))
    return FALSE;

  RemovePound(fullurl);

  if (!HandleToHeaders(handle,FHSIZE,metaptr,&headers)) {
    SafeFree(fullurl);
    return FALSE;
  }

  if ((strlen(fullurl)>1024) || (strlen(headers)>4096)) {
    SafeFree(fullurl);
    SafeFree(headers);
    return FALSE;
  }

  if (!ParseName(fullurl,host,&port,url)) {
    SafeFree(fullurl);
    SafeFree(headers);
    return FALSE;
  }

  sock = InitHttpSocket(SquidName,SquidPort); //host,port);
  if (sock==INVALID_SOCKET) {
    SafeFree(fullurl);
    SafeFree(headers);
    return FALSE;
  }

  fprintf(stderr,"head %s\n",fullurl);
  sprintf(msg,"HEAD %s HTTP/1.0\r\nProxy-Authorization: gecko\r\nCache-Control: only-if-cached\r\n%s\r\n",
    fullurl,
    headers);

  send(sock,msg,strlen(msg),0);

  fprintf(stderr,"%d\n",remain);

  while (((len=recv(sock,destptr,remain,0))>0) && (remain>0)) {
    remain-=len;
    total+=len;
  }

  closesocket(sock);
  SafeFree(fullurl);
  SafeFree(headers);

  if (strncmp(dest->Bytes+9,"504",3)!=0) {
    dest->ResultPtr = dest->Bytes+offset;
    dest->Count = MIN(MAX(0,total-offset),count);
    return TRUE;
  } else
    return FALSE;
}

#define D(x) ; //fprintf(stderr,"%d%s\n",tid,x)
#define D2(x) ; //fprintf(stderr,"--%s\n",x);

int SquidGetFull(pHttpBuffer dest)
{
  SOCKET sock;
  char host[1024];
  char url[1024];
  char msg[1024+4096+256];
  char buf[256];
  int port;
  int len;
  int total=0;

  char *fullurl, *headers;
  char *metadata;
  int inode;

  if (!HandleToUrl(dest->Handle,FHSIZE,&fullurl,&inode,&metadata)) {
    fprintf(stderr,"fullget htu fail\n");
    return FALSE;
  }

  RemovePound(fullurl);

  if (!HandleToHeaders(dest->Handle,FHSIZE,metadata,&headers)) {
    SafeFree(fullurl);
    fprintf(stderr,"fullget hth fail\n");
    return FALSE;
  }

  if ((strlen(fullurl)>1024) || (strlen(headers)>4096)) {
    SafeFree(fullurl);
    SafeFree(headers);
    fprintf(stderr,"fullget too long\n");
    return FALSE;
  }

  fprintf(stderr,"full get %s\n",fullurl);

  if (!ParseName(fullurl,host,&port,url)) {
    SafeFree(fullurl);
    SafeFree(headers);
    fprintf(stderr,"fullget parsename failed\n");
    return FALSE;
  }

  sock = InitHttpSocket(SquidName,SquidPort); //host,port);

  if (sock==INVALID_SOCKET) {
    SafeFree(fullurl);
    SafeFree(headers);
    fprintf(stderr,"fullget socket failed\n");
    return FALSE;
  }

  sprintf(msg,"GET %s HTTP/1.0\r\n%sProxy-Authorization: gecko\r\n\r\n",
    fullurl,
    headers);

  send(sock,msg,strlen(msg),0);

  dest->DoingHeaders=TRUE;

  while ((len=recv(sock,buf,sizeof(buf),0))>0) {
    total+=len;
    HttpBufCopyIn(dest,buf,len);
    COND_BROADCAST(dest->SignalDataReady);
//    Sleep(0);
  }

  fprintf(stderr,"received %d bytes %s/%s\n",total,host,url);

  closesocket(sock);

  SafeFree(fullurl);
  SafeFree(headers);

  return TRUE;
}

int CheckTestCase(fhandle handle, int offset, int count, pHttpData data, int wantheaders)
{
  char *fullurl = NULL;
  int inode;
  int bread;
  char *metadata;
  char *pound;
  FILE *f;

  if (!HandleToUrl(handle,FHSIZE,&fullurl,&inode,&metadata)) {
    fprintf(stderr,"testcase htu fail\n");
    return FALSE;
  }

  if (strnicmp(fullurl,"http://local_file/",18)!=0) {
    free(fullurl);
    return;
  }

  if ((pound=strchr(fullurl,'#'))!=NULL)
    *pound='\0';

  f=fopen(fullurl+18,"rb");
  if (f==NULL) {
    free(fullurl);
    return FALSE;
  }

  fprintf(stderr,"%d %d\n",offset,count);

  if (wantheaders) {
    if (offset!=0)
      data->Count=0;
    else {
      fseek(f,0,SEEK_END);
      data->ResultPtr=data->Bytes;
      sprintf(data->Bytes,"HTTP/1.0 200 OK\r\nContent-Length: %d\r\n\r\n",ftell(f));
      data->Count=strlen(data->Bytes);
    }
  } else {
    data->ResultPtr=data->Bytes;

    fseek(f,0,SEEK_END);
    if (offset>ftell(f))
      bread=0;
    else if (fseek(f,offset,SEEK_SET)==0)
      bread=fread(data->ResultPtr,1,count,f);
    else
      bread=0;

    if (bread<0)
      data->Count=0;
    else
      data->Count=bread;
  }

  fclose(f);

  free(fullurl);

  return TRUE;
}

int GetBytes(fhandle inhandle, int offset, int count, pHttpData data, int tid)
{
  pHttpBuffer HttpBuf;
  fhandle handle;

  // reset the type & inode bits so a .contents is the same as a .headers
  HandleCopy(handle,inhandle);
  InodeToHandle(handle,FHSIZE,0);
  SET_TYPE(handle,FLAG_CONTENTS);

  data->ResultPtr = NULL;
  data->Count = 0;

  if (CheckTestCase(handle,offset,count,data,FALSE))
    return TRUE;

  start:

  if ((Hash_FindEntry(&HashHttpBuf,handle)==NULL) && (offset<32768) && (SquidGetPartial(data,handle,offset,count)))
    return TRUE;

  MUTEX_LOCK(MutexHttpBuf);
  while ((HttpBuf = HttpBufFind(handle))==NULL)
    COND_WAIT(SignalBufferAvail,MutexHttpBuf);

  MUTEX_UNLOCK(MutexHttpBuf);

  while ((!HttpBufInRange(HttpBuf,&(HttpBuf->DataBuf),offset,count)) && (!HttpBuf->Completed)) {
    COND_WAIT(HttpBuf->SignalDataReady,HttpBuf->MutexBuffer);
  }

  if (HttpBufInRange(HttpBuf,&(HttpBuf->DataBuf),offset,count)) {
    HttpBufDec(HttpBuf);
    data->Count = CircularGet(data->Bytes,offset,count,&(HttpBuf->DataBuf));
    data->ResultPtr = data->Bytes;
    HttpBufUnlock(HttpBuf);
    return TRUE;
  } else {
    HttpBufDec(HttpBuf);
    HttpBufUnlock(HttpBuf);
    goto start;
  }
}

#define SKIPCR while ((x>data->ResultPtr) && (*x=='\r')) x--;
#define SKIPCRLF while ((x>data->ResultPtr) && ((*x=='\r') || (*x=='\n'))) x--;

void PadHeaders(pHttpData data)
{
  char *x = (data->ResultPtr+data->Count-1);
  int skip = (data->ResultPtr-data->Bytes);
  int i;

  if (data->Count<4)
    return;

  SKIPCR

  if (*x!='\n')
    return;

  SKIPCR

  if (*x!='\n')
    return;

  SKIPCRLF

  x++;
  x=stpcpy(x,"\r\nx-pad: ");
  for (i=0; i<240; i++)
    *(x++) = '*';

  *x = '\0';

  data->Count = (x-data->ResultPtr);
}

int GetHeaders(fhandle inhandle, int offset, int count, pHttpData data, int tid)
{
  pHttpBuffer HttpBuf;
  fhandle handle;

  // reset the type & inode bits so a .contents is the same as a .headers
  HandleCopy(handle,inhandle);
  InodeToHandle(handle,FHSIZE,0);
  SET_TYPE(handle,FLAG_CONTENTS);

  data->ResultPtr = NULL;
  data->Count = 0;

  if (CheckTestCase(handle,offset,count,data,TRUE))
    return TRUE;

  start:

  if ((Hash_FindEntry(&HashHttpBuf,handle)==NULL) && (SquidGetHead(data,handle,offset,count))) {
    PadHeaders(data);
    return TRUE;
  }

  MUTEX_LOCK(MutexHttpBuf);
  while ((HttpBuf = HttpBufFind(handle))==NULL)
    COND_WAIT(SignalBufferAvail,MutexHttpBuf);

  MUTEX_UNLOCK(MutexHttpBuf);

  while ((!HttpBufInRange(HttpBuf,&(HttpBuf->HeaderBuf),offset,count)) && (!HttpBuf->Completed))
    COND_WAIT(HttpBuf->SignalDataReady,HttpBuf->MutexBuffer);

  if (HttpBufInRange(HttpBuf,&(HttpBuf->HeaderBuf),offset,count)) {
    HttpBufDec(HttpBuf);
    data->Count = CircularGet(data->Bytes,offset,count,&(HttpBuf->HeaderBuf));
    data->ResultPtr = data->Bytes;
    HttpBuf->KeepaliveSeconds=3;
    HttpBufUnlock(HttpBuf);
    PadHeaders(data);
    return TRUE;
  } else {
    HttpBufDec(HttpBuf);
    HttpBufUnlock(HttpBuf);
    goto start;
  }
}

void WaitForRequest(pHttpBuffer buf)
{
  HttpBufLock(buf);
  while (!buf->RequestPending)
    COND_WAIT(buf->SignalRequestPending,buf->MutexBuffer);

  HttpBufUnlock(buf);
}

void SecondDelay(int seconds)
{
  struct timeval tv;

  tv.tv_sec=seconds;
  tv.tv_usec=0;

  select(0,NULL,NULL,NULL,&tv);
}

#ifdef WINDOWS
void BufferThread(void *arg)
#else
void *BufferThread(void *arg)
#endif
{
  pHttpBuffer buf = (pHttpBuffer) malloc(sizeof(struct THttpBuffer));

  COND_INIT(buf->SignalRefCountDec);
  COND_INIT(buf->SignalDataReady);
  COND_INIT(buf->SignalRequestPending);

  CircularInit(&buf->DataBuf,32768L*32L);
  CircularInit(&buf->HeaderBuf,8192);

  MUTEX_INIT(buf->MutexBuffer);

  HttpBufFree(buf);

  while (TRUE) {
    D2("wait");
    WaitForRequest(buf);

    ShowHandle(stderr,buf->Handle);

    D2("getfull");
    SquidGetFull(buf);

    HttpBufLock(buf);
    buf->Completed=TRUE;
    buf->KeepaliveSeconds=2;
    COND_BROADCAST(buf->SignalDataReady);
    if ((buf->DataBuf.Offset==0) && (buf->HeaderBuf.Offset==0)) {
      fprintf(stderr,"keepalive start\n");
      while (buf->KeepaliveSeconds>0) {
        HttpBufUnlock(buf);
        SecondDelay(1);
        HttpBufLock(buf);
        buf->KeepaliveSeconds--;
      }
      fprintf(stderr,"keepalive done\n");
    }
    HttpBufUnlock(buf);

    D2("delentry");
    MUTEX_LOCK(MutexHttpBuf);
    Hash_DeleteEntry(&HashHttpBuf,buf->HashEntry);
    buf->HashEntry=NULL;
    MUTEX_UNLOCK(MutexHttpBuf);

    D2("httpbuflock");
    HttpBufLock(buf);

    buf->Completed=TRUE;
    COND_BROADCAST(buf->SignalDataReady);

    while (buf->RefCount>0) {
//      fprintf(stderr,"--wait refcount(%d)\n",buf->RefCount);
      COND_WAIT(buf->SignalRefCountDec,buf->MutexBuffer);
    }

    buf->RequestPending=FALSE;

    D2("Unlock");
    HttpBufUnlock(buf);
    HttpBufFree(buf);
    COND_BROADCAST(SignalBufferAvail);
  }
}

void FragmentInit(char *squidname, int squidport)
{
  int i;

  MUTEX_INIT(MutexSocket);

  MUTEX_INIT(MutexHttpBuf);
  COND_INIT(SignalBufferAvail);

  Hash_InitTable(&HashHttpBuf,0,FHSIZE/4);

  for (i=0; i<WORKER_COUNT; i++) {
   CREATE_THREAD(ThreadID[i],BufferThread,(void*) i);
  }

  SquidName = SafeStrDup(squidname);
  SquidPort = squidport;
}

/*int OldGetBytes(fhandle handle; int offset; int count; pHttpData data)
{
  pHttpBuffer HttpBuf = HttpBufFind(handle);

  start:

  if ((HttpBuf==NULL) && (SquidGetPartial(handle,offset,count,data)))
    return TRUE;

  if (HttpBuf==NULL) {
    HttpBuf = BufferAlloc();
    if (HttpBuf==NULL) {
      COND_WAIT(SignalBufferAvail);
      goto start;
    } else {
      HttpBufStart(HttpBuf);
      goto start;
    }
  }

  if (HttpBuf!=NULL) {
    if (offset<HttpBuf->Offset) {
      HttpBufUnlock(HttpBuf);
      COND_WAIT(HttpBuf->Complete);
      goto start;
    } else if (offset+count<=HttpBuf->Offset+HttpBuf->Count) {
      HttpBufDec(HttpBuf);
      HttpBufCopyResult(data,HttpBuf,offset,count);
      HttpBufUnlock(HttpBuf);
      return TRUE;
    } else {
      HttpBufUnlock(HttpBuf);
      COND_WAIT(HttpBuf->DataReady);
      goto start;
    }
  } else {
    if (SquidGetPartial(handle,offset,count,data)) {
      HttpBufDec(HttpBuf);
      return TRUE;
    } else {
      HttpBuf = BufferAlloc();
      if (HttpBuf==NULL) {
        COND_WAIT(SignalBufferAvail);
        goto start;
      } else {
        HttpBufStart(HttpBuf);
        goto start;
      }
    }
  }
}*/








