#include "stdinc.h"
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#ifdef WINDOWS
#define NO_MMAP
#else
#include <sys/types.h>
#include <sys/mman.h>
#include <signal.h>
#include <mntent.h>
#include <pwd.h>
#endif

#include "buffer.h"
#include "md5file.h"

#define CONTENTS_NAME ".contents"
#define HEADERS_NAME ".headers"
#define HSET_DIRNAME "headersets"

#define FORWARDSLASH '/'
#define BACKSLASH '\\'

#include "thread.h"
#include "url.h"
#include "html.h"
#include "httpcon.h"

#define PROXY_PORT ProxyPort
#define PROXY_ADDR INADDR_ANY

#define MAX_THREADS 32
#define NUM_THREADS 10
#define SOCKET_QUEUE_SIZE 1024

#define PROGRAM "Gecko2 Client Proxy Version 1.1"

#define PRIVACY_NONE 0
#define PRIVACY_COOKIES_AND_AUTH 1
#define PRIVACY_ALL 2

#define DIAG_ALWAYS 0
#define DIAG_BASIC 1
#define DIAG_EXTRA 2
#define DIAG_VERBOSE 3

int NumThreads = NUM_THREADS;
int UID,GID;
int AllowInvalidate = TRUE;
int Diagnostics = DIAG_ALWAYS;
int DiagErrors = TRUE;
int UseHeaderSets=TRUE;
int Gecko2InvalidateCount=0;
int DeleteReferer=TRUE;
int PrivacyMode = PRIVACY_COOKIES_AND_AUTH;

struct passwd MyPasswordEntry;

pthread_key_t KeyThreadNum;

MUTEX_TYPE MutexMmap;

char PostServerName[256];
int PostServerPort;
char MountPoint[256];
int ProxyPort;
int IncludeServerInPath;
int UseDotContents, UseDotHeaders;

#define SENDFILE_OK 0
#define SENDFILE_MMAP_FAILED 1
#define SENDFILE_OPEN_FAILED 2
#define SENDFILE_STAT_FAILED 3
#define MAXERROR 3

char *ErrorTypes[] = {"Ok","sendfile mmap failed","sendfile open failed","sendfile stat failed"};

int NumError(int x, int y)
{
  return (x<<16) | y;
}

char *GetErrorType(int x)
{
  int errtype = x>>16;

  if ((errtype<0) || (errtype>MAXERROR))
    return "unknown error";
  else
    return ErrorTypes[errtype];
}

char *GetErrnoText(int x)
{
  int err = (x&0xFFFF);

  if ((err<0) || (err>=sys_nerr))
    return "unknown error";
  else
    return sys_errlist[errno];
}

void DiagMsg(int level, char *msg, ...)
{
    va_list args;

    char buf[4096];

    int threadnum = (int) pthread_getspecific(KeyThreadNum);

    va_start(args, msg);
    vsprintf(buf, msg, args);
    va_end(args);

    if (Diagnostics>=level)
      fprintf(stderr,"g2proxy(%d): %s\n",threadnum,buf);
}

void DiagError(char *msg, ...)
{
    va_list args;

    char buf[4096];
    int threadnum = (int) pthread_getspecific(KeyThreadNum);

    va_start(args, msg);
    vsprintf(buf, msg, args);
    va_end(args);

    if ((Diagnostics) || (DiagErrors)) {
      fprintf(stderr,"g2proxy(%d): %s\n",threadnum,buf);
      perror("error code=");
    }
}

void Error(char *s)
{
  printf("g2proxy: %s\n",s);
}

void LingeringClose(SOCKET sock)
{
  fd_set fds, fds_read, fds_err;
  char dummybuf[256];
  struct timeval tv;
  int select_rv;

  shutdown(sock,1);

  FD_ZERO(&fds);
  FD_SET(sock,&fds);

  do {
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    fds_read = fds;
    fds_err = fds;

    select_rv = select(sock+1,&fds_read,NULL,&fds_err,&tv);
  } while ((select_rv>0) && (!FD_ISSET(sock,&fds_err)) && (FD_ISSET(sock,&fds_read)) && (read(sock,dummybuf,sizeof(dummybuf))>0));

  closesocket(sock);
}

SOCKET InitSocket(void)
{
  struct sockaddr_in inaddr;
  SOCKET sock;
  int result;
  int reuseaddr = 1;

#ifdef WINDOWS
  WSADATA winsock_data;
  WORD version_required = 0x0101; /* Version 1.1 */

  if (WSAStartup (version_required, &winsock_data)!=0) {
    Error("unable to open winsock");
    return -1;
  }
#endif

  sock = socket(AF_INET, SOCK_STREAM, 0);

  if (sock==INVALID_SOCKET) {
    Error("unable to create socket");
    return -1;
  }

  setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&reuseaddr,sizeof(reuseaddr));

  inaddr.sin_family = AF_INET;
  inaddr.sin_port = htons(PROXY_PORT);
  inaddr.sin_addr.s_addr = htonl(PROXY_ADDR);

  if ((result=bind(sock,(struct sockaddr*) &inaddr,sizeof(inaddr)))!=0) {
    Error("unable to bind socket");
    closesocket(sock);
    return -1;
  }

  if ((result=listen(sock,SOMAXCONN))!=0) {
    Error("unable to listen");
    closesocket(sock);
    return -1;
  }

  return sock;
}

void TranslateName(char *in, char *out, char *hdrname)
{
  char host[2048], url[2048];
  int port;
  char last;
  char *urlp = url;

  ParseName(in,host,&port,url);

  TruncateAtFirstCharacter(url,'#');

  strcpy(out,MountPoint);

  if (UseHeaderSets) {
    AddSlash(out);
    strcat(out,HSET_DIRNAME);
    AddSlash(out);
    strcat(out,hdrname);
  }

  if (IncludeServerInPath) {
    AddSlash(out);
    strcat(out,host);
  }

  if (*url!='/')
    AddSlash(out);

  if (strstr(url,"//")!=NULL) {
    while (*urlp) {
      if ((*urlp=='/') && (last=='/'))
        strcat(out,".slash/.slash/");
      else
        strncat(out,urlp,1);
      last=*urlp;
      urlp++;
    }
  } else
    strcat(out,url);
}

void SendStatusLine(struct Buffer *buf, int code, char *msg)
{
  char tmp[256];

  sprintf(tmp,"HTTP/1.0 %d %s\r\n",code,msg);
  send(buf->Sock,tmp,strlen(tmp),0);

  DiagMsg(DIAG_BASIC,"status: %d %s",code,msg);
}

void SendCRLF(struct Buffer *buf)
{
  char *msg ="\r\n";

  send(buf->Sock,msg,strlen(msg),0);
}

/* Two methods of sending the file are included below:

   NO_MMAP: (Windows) Open the file and read/send in 1k chunks
   STANDARD: (Unix) Open & mmap the file and send it in one big chunk
*/

#ifndef NO_MMAP
void *mmap1(void *start,size_t length, int prot, int flags, int fd, off_t offset)
{
  void *result;

  MUTEX_LOCK(MutexMmap);
  result = mmap(start,length,prot,flags,fd,offset);
  MUTEX_UNLOCK(MutexMmap);

  return result;
}

int munmap1(void *start, size_t length)
{
  int result;

  MUTEX_LOCK(MutexMmap);
  result = munmap(start,length);
  MUTEX_UNLOCK(MutexMmap);

  return result;
}
#endif

void BeMoreSpecific(char *name)
{
  struct stat st;
  char *x = name;

  DiagMsg(DIAG_EXTRA,"stat() is reported below for each component of the path name");

  while (*x) {
    if (*x=='/') {
      *x='\0';
      fprintf(stderr,"  stat(%s) = %d\n",name,stat(name,&st));
      *x='/';
    }
    x++;
  }

  fprintf(stderr,"  stat(%s) = %d\n",name,stat(name,&st));
}

int SendFile(struct Buffer *buf, char *name)
{
  int pid = getpid();
  int handle;
  char b[32767];
  int bread;
#ifdef NO_MMAP
#else
  struct stat stat;
  char *mapaddress;
#endif

  DiagMsg(DIAG_BASIC,"sendfile(name=%s)",name);

  handle = open(name,O_BINARY | O_RDONLY);

  if (handle==-1) {
    int result = NumError(SENDFILE_OPEN_FAILED,errno);
    DiagError("sendfile unable to open %s",name);
    BeMoreSpecific(name);
    return result;
  }

#ifdef NO_MMAP
  while ((bread=read(handle,b,sizeof(b)))>0) {
    int i;

    for (i=0; i<bread; i++)
      fputc(b[i],stderr);

    if (send(buf->Sock,b,bread,0)==-1)
      DiagError("send failed on %s",name);
  }
#else
  if ((fstat(handle,&stat)!=0) || (stat.st_size<=0)) {
    int result=NumError(SENDFILE_STAT_FAILED,errno);
    close(handle);
    return result;
  }

  if ((mapaddress=mmap1(NULL,stat.st_size,PROT_READ,MAP_SHARED,handle,0))==-1) {
    int result=NumError(SENDFILE_MMAP_FAILED,errno);
    DiagError("mmap failed on %s",name);
    close(handle);
    return result;
  }

  if (send(buf->Sock,mapaddress,stat.st_size,0)==-1) {
    DiagError("send failed on %s (addr=%d,size=%d,tid=%d)",name,mapaddress,stat.st_size,pid);

    bread=read(handle,b,sizeof(b));
    if (bread==-1)
      perror("read failed");
    else {
      fprintf(stderr,"read result = %d\n",bread);
      if (send(buf->Sock,b,bread,0)==-1)
        perror("send of read failed");
      else
        fprintf(stderr,"send succeeded\n");
    }
  }
  munmap1(mapaddress,stat.st_size);
#endif

  close(handle);

  return SENDFILE_OK;
}

void ProcessPost(struct Buffer *buf)
{
  fd_set readfds;
  char databuf[4096];
  int bread;
  int highsock;
  int count;
  int done1,done2;

  DiagMsg(DIAG_BASIC,"**** Handling Post: %s",buf->LineStart);

  buf->OutSocket = InitHttpSocket(PostServerName,PostServerPort);

  if (buf->OutSocket!=-1) {
    send(buf->OutSocket,buf->LineStart,strlen(buf->LineStart),0);
    send(buf->OutSocket,"\r\n",2,0);
    BufferDump(buf);

    done1=FALSE;
    done2=FALSE;
    while ((!done1) || (!done2)) {
      FD_ZERO(&readfds);
      FD_SET(buf->Sock,&readfds);
      FD_SET(buf->OutSocket,&readfds);

      highsock=buf->Sock+1;
      if (buf->OutSocket>=highsock)
        highsock=buf->OutSocket+1;

      count=select(highsock,&readfds,NULL,NULL,0);

      if (count>0) {
        if ((!done1) && (FD_ISSET(buf->Sock,&readfds))) {
          bread=recv(buf->Sock,databuf,sizeof(databuf),0);
          DiagMsg(DIAG_EXTRA,"%d byes in from client",bread);
          if (bread==0) {
            done1=TRUE;
            DiagMsg(DIAG_EXTRA,"client closed connection");
            shutdown(buf->OutSocket,1);
          } else if (bread<0) {
            DiagMsg(DIAG_EXTRA,"client errored connection");
            done1=TRUE;
            done2=TRUE;
          } else
            send(buf->OutSocket,databuf,bread,0);
        }
        if ((!done2) && (FD_ISSET(buf->OutSocket,&readfds))) {
          bread=recv(buf->OutSocket,databuf,sizeof(databuf),0);
          DiagMsg(DIAG_EXTRA,"%d bytes in from server",bread);
          if (bread==0) {
            done2=TRUE;
            DiagMsg(DIAG_EXTRA,"server closed connection");
            shutdown(buf->Sock,1);
          } else if (bread<0) {
            DiagMsg(DIAG_EXTRA,"server errored connection");
            done1=TRUE;
            done2=TRUE;
          } else
            send(buf->Sock,databuf,bread,0);
        }
      }
    }

    closesocket(buf->OutSocket);
    buf->OutSocket=INVALID_SOCKET;
    buf->IsEof = TRUE;
  } else
    DiagError("Unable to open socket to post server");

  DiagMsg(DIAG_BASIC,"post completed");
}

void UpdateHeaderSet(char *hdrbuf, char *hdrname, int shouldbeprivate)
{
  char genname[1024];
  char hsetname[1024];
  int result;
  FILE *fgen;
#ifndef WINDOWS
  struct stat st;
#endif

  if (!UseHeaderSets)
    return;

  DiagMsg(DIAG_VERBOSE,"Processing header set");

  strcpy(hsetname,MountPoint);
  AddSlash(hsetname);
  strcat(hsetname,HSET_DIRNAME);
  AddSlash(hsetname);
  strcat(hsetname,hdrname);

  strcpy(genname,MountPoint);
  AddSlash(genname);
  strcat(genname,".generator");

  #ifndef WINDOWS
  result = (stat(hsetname,&st)==0);
  #else
  result = FALSE;
  #endif

  if (result) {
    DiagMsg(DIAG_BASIC,"headerset %s exists",hdrname);
    DiagMsg(DIAG_EXTRA,"... contents: %s",hdrbuf);
  } else {
    DiagMsg(DIAG_BASIC,"creating headerset %s",hdrname);
    DiagMsg(DIAG_EXTRA,"... contents: %s",hdrbuf);

    fgen = fopen(genname,"wt");

    if (fgen==NULL) {
      DiagError("GENERATOR FAILURE (RETRY): unable to write to %s",genname);
      DiagMsg(DIAG_EXTRA,"(Explanation: g2proxy could not open .generator for writing. g2 proxy will try it again)");
      BeMoreSpecific(genname);
      fgen = fopen(genname,"wt");
    }

    if (fgen!=NULL) {
      fprintf(fgen,"%s <RESET> %s",hdrname,hdrbuf);
      fclose(fgen);

      DiagMsg(DIAG_EXTRA,"mkdir(%s)=%d",hsetname,mkdir(hsetname,S_IRUSR+S_IWUSR+S_IXUSR));

      if (shouldbeprivate) {
        DiagMsg(DIAG_EXTRA,"headerset will be configured as private");
        chown(hsetname,UID,GID);
        chmod(hsetname,S_IRUSR+S_IWUSR+S_IXUSR);
      } else
        DiagMsg(DIAG_EXTRA,"headerset will be configured as public");
    } else {
      DiagError("GENERATOR FAILURE (FATAL): unable to write to %s",genname);
      DiagMsg(DIAG_EXTRA,"(Explanation: g2proxy tried to open .generator twice and failed.)");
      BeMoreSpecific(genname);
    }
  }
}

void ProcessHttpHeaders(char *url,
			struct HttpHeaders *hdrs,
                        int *nocache,
                        int *ims,
                        char *hdrbuf,
                        char *hdrname,
                        int *shouldbeprivate)
{
  char hexmap[16] = "0123456789ABCDEF";
  int i;
  char *hdrptr = hdrbuf;
  md5file md5;
  unsigned char md5result[MD5FILE_HASHLEN];
  char hostname[1024];

  GetHostNameFromUrl(hostname,url);

  (*nocache)=FALSE;
  (*ims)=FALSE;
  (*shouldbeprivate)=(PrivacyMode==PRIVACY_ALL);

  for (i=0; i<hdrs->Count; i++) {
    char *tag = hdrs->Tag[i];
    char *data = hdrs->Data[i];

    if (stricmp(tag,"if-modified-since")==0) {
      (*nocache)=TRUE;
      (*ims)=TRUE;
      DiagMsg(DIAG_VERBOSE,"received if-modified-since header");
    } else if ((stricmp(tag,"pragma")==0) && (data!=NULL) && (stricmp(data,"no-cache")==0)) {
      (*nocache)=TRUE;
      DiagMsg(DIAG_VERBOSE,"received pragma: no-cache header");
    } else if ((DeleteReferer) && (stricmp(tag,"referer")==0)) {
      // do nothing
      DiagMsg(DIAG_VERBOSE,"received and ignored referer header");
    } else if (((stricmp(tag,"host")==0) && (stricmp(data,hostname)==0))) {
      // do nothing
      DiagMsg(DIAG_VERBOSE,"received and ignored host header");
    } else if (stricmp(tag,"Proxy-Connection")==0) {
      // do nothing
      DiagMsg(DIAG_VERBOSE,"received and ignored Proxy-Connection header");
    } else if ((tag!=NULL) && (data!=NULL))
      hdrptr=hdrptr+sprintf(hdrptr,"%s: %s\r\n",tag,data);
      DiagMsg(DIAG_VERBOSE,"appended header: %s: %s",tag,data);

    if ((PrivacyMode==PRIVACY_COOKIES_AND_AUTH) &&
        ((stricmp(tag,"www-authenticate")==0) || (stricmp(tag,"cookie")==0))) {
      (*shouldbeprivate)=TRUE;
      DiagMsg(DIAG_VERBOSE,"headerset should be private (cookies or auth)");
    }
  }

//  *(hdrname++)='#';

  if (*shouldbeprivate) {
    hdrptr=hdrptr+sprintf(hdrptr,"PrivateUser: %s\r\n",MyPasswordEntry.pw_name);
    DiagMsg(DIAG_VERBOSE,"appended header: PrivateUser: %s",MyPasswordEntry.pw_name);
  }

  md5file_clear(&md5);
  md5file_addn(&md5,hdrbuf,strlen(hdrbuf));
  md5file_hash(&md5,md5result);

  *(hdrname++)='#';
  for (i=0; i<MD5FILE_HASHLEN; i++) {
    *(hdrname++) = hexmap[md5result[i]>>4];
    *(hdrname++) = hexmap[md5result[i]&0x0F];
  }
  *(hdrname)='\0';
}

void CheckInvalidChars(char *name)
{
  char *x;
  int pos=0;

  for (x=name; *x; x++) {
    pos++;
    if ((*x<32) || (*x>=127))
      DiagMsg(DIAG_ALWAYS,"Possible invalid character. pos=%d, value=%d",pos,*x);
  }
}

void ProcessGet(struct Buffer *buf)
{
  char basename[2048],name[2048],headername[2048],req[2048];
  char hdrname[256], hdrbuf[4096];
  char *x;
  int didsendheaders;
  struct HttpHeaders hdrout;
  char *hash;
  int sendresult;

  int nocache,ims;
  int shouldbeprivate=FALSE;

  DiagMsg(DIAG_BASIC,"**** Handling Get: %s",buf->LineStart);

  strcpy(req,buf->LineStart+4);
  TruncateAtFirstCharacter(req,' ');

  InitHttpHeaders(&hdrout);

  DiagMsg(DIAG_VERBOSE,"GetHttpHeaders");
  GetHttpHeaders(buf,HTTP_STATE_HEADERS,&hdrout);

  DiagMsg(DIAG_VERBOSE,"ProcessHttpHeaders");
  ProcessHttpHeaders(req,&hdrout,&nocache,&ims,hdrbuf,hdrname,&shouldbeprivate);

  DiagMsg(DIAG_VERBOSE,"UpdateHeaderSet");
  UpdateHeaderSet(hdrbuf,hdrname,shouldbeprivate);

  DiagMsg(DIAG_VERBOSE,"TranslateName");
  TranslateName(req,basename,hdrname);

#ifdef WINDOWS
  for (x = basename; *x; x++)
    if (*x==FORWARDSLASH)
      *x=BACKSLASH;
#endif
  if (*(basename+strlen(basename)-1)=='/') {
    strcat(basename,".slash");
    DiagMsg(DIAG_VERBOSE,"appended .slash to basename");
  }

  strcpy(name,basename);

  if (UseDotContents) {
    AddSlash(name);
    strcat(name,CONTENTS_NAME);
  }

  CheckInvalidChars(name);

  if (nocache)
    Gecko2InvalidateCount++;

  if (Gecko2InvalidateCount>0)
    sprintf(name+strlen(name),"#%d",Gecko2InvalidateCount);

  if (UseDotHeaders) {
    strcpy(headername,basename);
    AddSlash(headername);
    strcat(headername,HEADERS_NAME);

    if (Gecko2InvalidateCount>0)
      sprintf(headername+strlen(headername),"#%d",Gecko2InvalidateCount);

    didsendheaders = (SendFile(buf,headername)==SENDFILE_OK);
  } else
    didsendheaders = FALSE;

  if (didsendheaders)
    SendCRLF(buf);
  else
    SendStatusLine(buf,200,"Document Follows");

  SendCRLF(buf);

  sendresult = SendFile(buf,name);
  if (sendresult!=0) {
    char badmsg[2048];

    DiagMsg(DIAG_BASIC,"Unable to retrieve nfs file; sending bad response");

    sprintf(badmsg,
            "<html><head><title>Error</title></head><body>Unable to retrieve the NFS file (%s) corresponding to your request<P>Specific error: %s/%s</body></html>\n",
            name,
            GetErrorType(sendresult),
            GetErrnoText(sendresult));

    send(buf->Sock,badmsg,strlen(badmsg),0);
  }
}

void ProcessRequest(SOCKET sock)
{
  struct Buffer *buf = BufferCreate();

  buf->Sock = sock;

  if (BufferGetLine(buf)) {
    if (strnicmp(buf->LineStart,"GET ",4)==0)
      ProcessGet(buf);
    else if (strnicmp(buf->LineStart,"POST ",5)==0) {
      ProcessPost(buf);
    } else  {
      DiagMsg(DIAG_ALWAYS,"command not implemented: %s\n",buf->LineStart);
      SendStatusLine(buf,500,"proxy does not implement that command!");
      SendCRLF(buf);
    }
  } else
    DiagError("BufferGetLine failed\n");

  closesocket(sock);
}

typedef struct SocketNode *pSocketNode;
struct SocketNode {
  pSocketNode Next;
  SOCKET Socket;
};

THREAD_TYPE ThreadID[MAX_THREADS];
SEMAPHORE_TYPE SemaphoreRequestWaiting;
MUTEX_TYPE MutexSocketQueue;
pSocketNode SocketQueue, SocketQueueFree, SocketQueueLast;
int SocketQueueCount = 0;

SOCKET GetSocketFromQueue(void)
{
  SOCKET s;
  pSocketNode newnode;

  MUTEX_LOCK(MutexSocketQueue);

  if (SocketQueue==NULL) {
    DiagMsg(DIAG_ALWAYS,"out of socket queue entries!\n");
    exit(1);
  }

  newnode = SocketQueue;
  SocketQueue=SocketQueue->Next;

  newnode->Next = SocketQueueFree;
  SocketQueueFree=newnode;

  s=newnode->Socket;

  SocketQueueCount--;

  MUTEX_UNLOCK(MutexSocketQueue);

  return s;
}

int AddSocketToQueue(SOCKET s)
{
  pSocketNode newnode;

  MUTEX_LOCK(MutexSocketQueue);

  if (SocketQueueFree==NULL) {
    char *msg = "HTTP/1.0 503 Server is overloaded!\r\n\r\n";
    int len;

    MUTEX_UNLOCK(MutexSocketQueue);

    len=strlen(msg);
    send(s,msg,len,0);
    LingeringClose(s);

    return FALSE;
  }

  newnode = SocketQueueFree;
  SocketQueueFree=SocketQueueFree->Next;

  newnode->Next = NULL;

  if (SocketQueue!=NULL) {
    SocketQueueLast->Next = newnode;
    SocketQueueLast = newnode;
  } else {
    SocketQueue = newnode;
    SocketQueueLast = newnode;
  }

  newnode->Socket = s;

  SocketQueueCount++;

  MUTEX_UNLOCK(MutexSocketQueue);

  return TRUE;
}

#ifdef WINDOWS
_USERENTRY WorkerThread(LPVOID arg)
#else
void *WorkerThread(void *arg)
#endif
{
  SOCKET asock;
  int threadnumber = (int) arg;

  pthread_setspecific(KeyThreadNum,(void*)threadnumber);

  while (TRUE) {
    WAIT_SEMAPHORE(SemaphoreRequestWaiting);

    DiagMsg(DIAG_VERBOSE,"thread woke up");

    asock = GetSocketFromQueue();

    DiagMsg(DIAG_VERBOSE,"received socket %d from queue",asock);

    ProcessRequest(asock);
  }
}

void CreateWorkerThreads(void)
{
  int i;

  pthread_key_create(&KeyThreadNum,NULL);
  pthread_setspecific(KeyThreadNum,(void*)-1);

  CREATE_SEMAPHORE(SemaphoreRequestWaiting);

  for (i=0; i<NumThreads; i++)
    CREATE_THREAD(ThreadID[i],WorkerThread,(void*) i);

  DiagMsg(DIAG_ALWAYS,"%d worker threads created\n");
}

void CreateSocketQueue(void)
{
  int i;

  SocketQueue = NULL;
  SocketQueueFree = NULL;

  for (i=0; i<SOCKET_QUEUE_SIZE; i++) {
    struct SocketNode *newnode = (struct SocketNode *) malloc(sizeof(struct SocketNode));

    newnode->Next = SocketQueueFree;
    newnode->Socket = INVALID_SOCKET;
    SocketQueueFree = newnode;
  }
}

static struct { char *name, *use ; } Args[] = {
  {"-h",        "Print usage information"},
  {"-m",        "Specify mount point -m xxx (default is /mnt/web)"},
  {"-n",        "Do not include server name in path"},
  {"-c",	"Do not use .contents"},
  {"-r",	"Do not use .headers"},
  {"-pn",	"Post Server Name -pn xxx (default is gecko.cs.arizona.edu)"},
  {"-pp",	"Post Server Port -pp xxx (default is 3128)"},
  {"-p",	"Specify proxy port -p xxx (Default is 8080)"},
  {"-nohs",	"Do not use Header Sets"},
  {"-d1",	"Basic Diagnostics (minimal)"},
  {"-d2",	"Extended Diagnostics (some)"},
  {"-d3",	"Verbose Diagnostics (a lot)"},
  {"-noinval",	"Do not invalidate (delete) objects"},
  {"-nt",	"Number of threads -nt xxx (default is 10)"},
  {"-pAll",	"Make all requests private"},
  {"-pNone",	"Make all requests public"},
  {"-pAuth",	"Make only authentication and cookies private (default)"},
  {NULL, NULL}
  } ;

static void usage( char *name ) {
  int f ;

  printf( "usage: %s [flags]\n", name ) ;
  printf( "The following flags are recognized:\n" ) ;
  for( f = 0 ; Args[f].name != NULL ; f++ )
    printf( "flag: %-5s  meaning: %s\n", Args[f].name, Args[f].use ) ;
  }


#define MOREARGS (*argv[++a]!='\0')

int GetMountPoint(char *mountpoint)
{
  FILE *f;

  f=setmntent("/etc/mtab","rt");

  if (f!=NULL) {
    struct mntent *mnt;
    struct mntent static_mnt;
    char buf[1024];

    while ((mnt=getmntent_r(f,&static_mnt,buf,sizeof(buf)))!=NULL) 
      if ((strcmp(mnt->mnt_type,"nfs")==0) && (strstr(mnt->mnt_fsname,":web")!=NULL)) {
        strcpy(mountpoint,mnt->mnt_dir);
        endmntent(f);
        return TRUE;
      }

    endmntent(f);
  }

  return FALSE;
}

static void ProcessCommandLine( int argc, char *argv[] ) {
  int a, f ;

  strcpy(MountPoint,"/mnt/web");
  if (GetMountPoint(MountPoint))
    fprintf(stdout,"g2proxy: found existing mount point at %s\n",MountPoint);
  else
    fprintf(stderr,"g2proxy: WARNING: No existing gecko mount detected!\n");

  IncludeServerInPath = TRUE;
  UseDotContents = TRUE;
  UseDotHeaders = TRUE;
  strcpy(PostServerName,"gecko.cs.arizona.edu");
  PostServerPort=3128;
  ProxyPort=8080;

  for( a = 1 ; a < argc ; a++ ) {
    char *cmdname = argv[a];

    for( f = 0 ; Args[f].name != NULL ; f++ )
      if( stricmp(Args[f].name, cmdname) == 0 ) break ;

    // extra check for --help
    if (stricmp(cmdname,"--help")==0)
      f=0;

    switch( f ) {
      case 0: usage("g2proxy");
      	      exit(0);
              break; /* usage information */

      case 1: if (MOREARGS)
                strcpy(MountPoint,argv[a]);
	      break;

      case 2: IncludeServerInPath = FALSE;
	      break;

      case 3: UseDotContents = FALSE;
      	      break;

      case 4: UseDotHeaders = FALSE;
      	      break;

      case 5: if (MOREARGS)
                strcpy(PostServerName,argv[a]);
              break;

      case 6: if (MOREARGS)
                PostServerPort = atoi(argv[a]);
              break;

      case 7: if (MOREARGS)
      		ProxyPort=atoi(argv[a]);
              break;

      case 8: UseHeaderSets=FALSE;
      	      break;

      case 9: Diagnostics=DIAG_BASIC;
              break;

      case 10: Diagnostics=DIAG_EXTRA;
      	       break;

      case 11: Diagnostics=DIAG_VERBOSE;
      	       break;

      case 12: AllowInvalidate=FALSE;
      	       break;

      case 13: if (MOREARGS)
      		 NumThreads = atoi(argv[a]);
               break;

      case 14: PrivacyMode=PRIVACY_ALL;
      	       break;

      case 15: PrivacyMode=PRIVACY_NONE;
      	       break;

      case 16: PrivacyMode=PRIVACY_COOKIES_AND_AUTH;
      	       break;
    }
  }

  DiagMsg(DIAG_ALWAYS,"will listen on TCP socket %d\n",ProxyPort);
}

int main(int argc, char **argv)
{
  SOCKET sock;
  SOCKET asock;
  struct sockaddr_in remote_addr;
  int addrlen = sizeof(remote_addr);

  #ifdef NEED_PTHREAD_INIT
  pthread_init();
  #endif

  #ifdef WINDOWS
  UID=500;
  GID=500;
  #else
  CREATE_MUTEX(MutexMmap);
  signal(SIGPIPE,SIG_IGN);
  UID=getuid();
  GID=getgid();
  MyPasswordEntry=*getpwuid(UID);
  #endif

  fprintf(stdout,"\n%s\n\n",PROGRAM);

  ProcessCommandLine(argc,argv);

  sock = InitSocket();

  CreateSocketQueue();

  CreateWorkerThreads();

  fprintf(stdout,"g2proxy: online and waiting for requests...\n");

  while (TRUE) {
    while ((asock = accept(sock,(struct sockaddr*) &remote_addr,&addrlen))!=INVALID_SOCKET) {
      DiagMsg(DIAG_VERBOSE,"incoming request accepted socket=%d",asock);
      if (AddSocketToQueue(asock)) {
        DiagMsg(DIAG_VERBOSE,"added socket %d to queue",asock);
        SIGNAL_SEMAPHORE(SemaphoreRequestWaiting);
      }
    }
  }
}




