#include "stdinc.h"
#include <sys/mman.h>
#include <assert.h>
#include <pwd.h>
#include "nfs.h"
#include "thread.h"
#include "myrun.h"
#include "preferences.h"
#include "servmain.h"
#include "handle.h"
#include "fragment.h"
#include "dircache.h"
#include "stats.h"
#include "attr.h"
#include "geckonfs.h"
#include "buffer.h"
#include "html.h"
#include "parse.h"
#include "g2hset.h"
#include "compress.h"
#include "lru.h"
#include "inode.h"

fhandle RootHandle;
fhandle HsetHandle;
fhandle GeneratorHandle;
struct passwd NobodyPasswordEntry;

SOCKET SocketForHttp = INVALID_SOCKET;

void HandleHttpRequest(SOCKET s)
{
}

void debug(char *x, char *y, fhandle *handle)
{
  FILE *f;

  if (!Preferences.Debug)
    return;

  f = stderr;

  if (f != (FILE *)NULL) {

    fprintf(f,"DEBUG: %d: ",getpid());

    if (x!=NULL)
      fprintf(f, "%s ", x);

    if (y!=NULL)
      fprintf(f,"%s ",y);

    if (handle!=NULL)
      ShowHandle(f,*handle);
    else
      fprintf(f,"\r\n");

    if (f!=stderr) fclose(f);
  }
}

// return FALSE if
//   name is a headerset (i.e. #001)
//   name is a symbolic link (i.e. .link)
int UserCanAddName(char *name)
{
  return ((*name!='#') && (strncmp(name,".link",5)!=0));
}

char *ConcatUrl(char *first, char *second)
{
  char *url = (char*) SafeMalloc(strlen(first)+strlen(second)+2);
  char *tmp;

  if (strcmp(second,"..")==0) {
    strcpy(url,first);
    tmp=strrchr(url,'/');
    if (tmp!=NULL)
      *tmp='\0';
  } else if (strcmp(second,".slash")==0) {
    strcpy(url,first);
    strcat(url,"/");
//    AddSlash(url);
  } else if (*second=='.') {
    strcpy(url,first);
    if (strchr(second,'#')!=NULL)     // for refresh, add the #xx to the
      strcat(url,strchr(second,'#')); // end of the URL
  } else {
    strcpy(url,first);
    AddSlash(url);
    strcat(url,second);
  }

  return url;
}

void HeaderLoadFromFile(char **dest, char *fn)
{
  FILE *f;
  char buf[4096+1];
  char buf2[1];

  buf2[0]='\0';

  memset(buf,0,sizeof(buf));

  f = fopen(fn,"rb");
  if (f!=NULL) {
    fread(buf,1,4096,f);
    fclose(f);
  }

  HsetUpdate(dest,buf,buf2);
}

void CreateRoot(void)
{
  char *metaptr;
  char *headers;

  HeaderLoadFromFile(&headers,"root.hset");

  UrlToHandle(ROOTNAME,ROOTINODE,RootHandle,FHSIZE,&metaptr);
  HeadersToHandle(headers,RootHandle,FHSIZE,metaptr);
  SET_TYPE(RootHandle,FLAG_ROOT);
  InodeCounter++;

  UrlToHandle(ROOTNAME,InodeCreate(ROOTINODE,FALSE),GeneratorHandle,FHSIZE,&metaptr);
  HeadersToHandle(headers,GeneratorHandle,FHSIZE,metaptr);
  SET_TYPE(GeneratorHandle,FLAG_GENERATOR);
  InodeCounter++;

  UrlToHandle(ROOTNAME,InodeCreate(ROOTINODE,FALSE),HsetHandle,FHSIZE,&metaptr);
  HeadersToHandle(headers,HsetHandle,FHSIZE,metaptr);
  SET_TYPE(RootHandle,FLAG_HARDLINK);
  InodeCounter++;

  SafeFree(headers);

  DirAddChild(RootHandle,GeneratorHandle,".generator",FALSE);
  DirAddChild(RootHandle,HsetHandle,HSET_DIRNAME,FALSE);

  ShowHandle(stderr,RootHandle);
}

void ServerInit(void)
{
  struct passwd *nobody_pw = getpwnam("nobody");

  if (nobody_pw==NULL) {
    fprintf(stderr,"'nobody' not found -- using uid=100 & gid=100\n");
    NobodyPasswordEntry.pw_uid=100;
    NobodyPasswordEntry.pw_gid=100;
  } else
    NobodyPasswordEntry = *nobody_pw;

  if (*Preferences.ProxyName=='\0') {
    strcpy(Preferences.ProxyName,"198.0.0.100");
    Preferences.ProxyPort=3128;
  }

  InitStatistics();
  CompressInit();
  InodeInit();
  UrlTableInit();
  DirInit();
  FragmentInit(Preferences.ProxyName,Preferences.ProxyPort);
  AttrInit();
  ParseInitModule();

  CreateRoot();

  fprintf(stderr,"server initialized\n");
}

/* -------------- mount procs follow -------------- */

fhstatus *
mountproc_mnt_1(dirpath *argp, struct svc_req *rqstp)
{
  fhstatus *result = (fhstatus*) AllocResult(sizeof(fhstatus));
  char *path = *argp;
  pStatInfo stat;

  debug("mnt",path,NULL);

  stat=StartStatistics("mount");

  while (*path=='/')
    path++;

  fprintf(stderr,"%s\n",path);

  if (strcmp(path,"web")==0) {
    ShowHandle(stderr,RootHandle);
    HandleCopy(result->fhstatus_u.directory,RootHandle);
    result->status = MNT3_OK;
  } else
    result->status = MNT3ERR_NOENT;

  RecordStatistics("mount",stat);

  debug("mnt complete",NULL,NULL);

  return (result);
}

exports *
mountproc_export_1(void *argp, struct svc_req *rqstp)
{
  exports *result = (exports*) AllocResult(sizeof(exports));
  static exports exp = NULL;
  pStatInfo stat;

  debug("export",NULL,NULL);

  stat=StartStatistics("export");

  if (exp==NULL) {
    exp = (exports) malloc(sizeof(exportnode));

    exp->ex_dir = strdup("web");
    exp->ex_groups = (groups) malloc(sizeof(groupnode));
    exp->ex_next = NULL;

    exp->ex_groups->gr_name = "*";
    exp->ex_groups->gr_next = NULL;
  }

  *result = exp;

  RecordStatistics("export",stat);

  return (result);
}

/* ----------------- nfs procedures ---------------- */

attrstat *
nfsproc_getattr_2(fhandleargs *argp, struct svc_req *rqstp)
{
  attrstat *result = (attrstat*) AllocResult(sizeof(attrstat));
  pStatInfo stat;

  debug("nfs_getattr",NULL,(fhandle*) &argp->data);

  stat=StartStatistics("nfs_getattr");

  AttrLookupConsistant(argp->data,&result->attrstat_u.attributes);
  result->status = NFS_OK;

  debug("nfs_getattr complete",NULL,NULL);

  RecordStatistics("getattr",stat);

  return (result);
}

attrstat*
nfsproc_setattr_2(sattrargs *argp, struct svc_req *rqstp)
{
  attrstat *result = (attrstat*) AllocResult(sizeof(attrstat));
  pStatInfo stat;

  debug("nfs_setattr",NULL,(fhandle*) &argp->file);

  stat=StartStatistics("nfs_setattr");

  AttrLookupConsistant(argp->file,&result->attrstat_u.attributes);

  fprintf(stderr,"  mode=%x, uid=%x, gid=%x\n",argp->attributes.mode,argp->attributes.uid,argp->attributes.gid);

  if ((result->attrstat_u.attributes.uid==NobodyPasswordEntry.pw_uid) ||
     (result->attrstat_u.attributes.uid==argp->attributes.uid) ||

     ((result->attrstat_u.attributes.mode|argp->attributes.mode==result->attrstat_u.attributes.mode) &&
     (argp->attributes.uid=0xFFFFFFFF) &&
     (argp->attributes.gid=0xFFFFFFFF))) {
    AttrSetPermissions(argp->file,&argp->attributes);
    AttrIncrementTime(argp->file);
  }

  debug("nfs_setattr complete",NULL,NULL);

  RecordStatistics("setattr",stat);

  return (result);
}

diropres *nfsproc_lookup_2(diropargs *argp, struct svc_req *rqstp)
{
  diropres *result = (diropres*) AllocResult(sizeof(diropres));
  pStatInfo stat;
  char *url;
  char *metaptrparent, *metaptrchild;
  int parentinode;
  int nodircache;

  assert(argp);
  assert(argp->name);
  assert(argp->dir);

  stat=StartStatistics("nfs_lookup");

  debug("nfs_lookup",argp->name,&argp->dir);

  if (strcmp(argp->name,"quit_the_server_now")==0)
    exit(1);

  if ((*argp->name=='.') && (strncmp(argp->name,".slash",6)!=0))
    DirMakeConsist(argp->dir);

  if (DirLookup(argp->dir,argp->name,result->diropres_u.diropok.file)) {
    DirSetParent(result->diropres_u.diropok.file,argp->dir,argp->name);
    AttrLookupConsistant(result->diropres_u.diropok.file,&result->diropres_u.diropok.attributes);
    result->status = NFS_OK;
  } else if ((UserCanAddName(argp->name)) && (HandleToUrl(argp->dir,FHSIZE,&url,&parentinode,&metaptrparent))) {
    char *childurl = ConcatUrl(url,argp->name);
    char *headers;

    UrlToHandle(childurl,InodeCreate(parentinode,TRUE),result->diropres_u.diropok.file,FHSIZE,&metaptrchild);
    HandleInitFileType(result->diropres_u.diropok.file,argp->name);

    HandleToHeaders(argp->dir,FHSIZE,metaptrparent,&headers);
    HeadersToHandle(headers,result->diropres_u.diropok.file,FHSIZE,metaptrchild);
    SafeFree(headers);

    if (strchr(argp->name,'#')==NULL) {
      DirAddChild(argp->dir,result->diropres_u.diropok.file,argp->name,FALSE);
      AttrIncrementTime(argp->dir);
    }

    AttrLookupConsistant(result->diropres_u.diropok.file,&result->diropres_u.diropok.attributes);
    result->status = NFS_OK;
    SafeFree(url);
    SafeFree(childurl);

    DirSetParent(result->diropres_u.diropok.file,argp->dir,argp->name);
  } else
    result->status = NFSERR_NOENT;

  debug("nfs_lookup complete",NULL,NULL);

  RecordStatistics("nfs_lookup",stat);

  return (result);
}

void MorphHeaders(char *update, char *oldheaders, char**newheaders)
{
  char *colon = strchr(update,':');
  int taglen = colon-update+1;
  int oldlen = (oldheaders==NULL) ? 0 : strlen(oldheaders);

  (*newheaders) = (char*) SafeMalloc(oldlen+strlen(update)+6);

  if (oldlen==0)
    (**newheaders)='\0';
  else
    strcpy(*newheaders,oldheaders);

  strcat(*newheaders,update);
  strcat(*newheaders,"\r\n");
}

attrstat *nfsproc_write_2(writeargs *argp, struct svc_req *rqstp)
{
  attrstat *result = (attrstat*) AllocResult(sizeof(attrstat));
  pStatInfo stat;
  fhandle dest,hshandle,oldhandle;
  char *hsdata, *hsname;
  char *hscopy;
  char *url;
  int inode;
  char *oldheaders, *newheaders;
  char *metaptr;
  int reset = FALSE;

  stat=StartStatistics("nfs_write");

  debug("nfs_write","",&argp->file);

  if (HandleGetType(argp->file)!=FLAG_GENERATOR) {
    result->status=NFSERR_PERM;
    RecordStatistics("nfs_write_notgenerator",stat);
    return result;
  }

  // duplicate the nfs data so we can null-terminate it
  hscopy = (char*) SafeMalloc(argp->data.nfsdata_len+1);
  memcpy(hscopy,argp->data.nfsdata_val,argp->data.nfsdata_len);
  *(hscopy+argp->data.nfsdata_len)='\0';

  // find a space to delimit headerset name from data
  hsdata=strchr(hscopy,' ');
  if (hsdata==NULL) {
    SafeFree(hscopy);
    result->status=NFSERR_IO;
    RecordStatistics("nfs_write_bogus",stat);
    return result;
  }

  // null terminate the name and skip data over any extra whitespace
  *(hsdata++)='\0';
  while (isspace(*hsdata))
    hsdata++;

  // see if there is a <Reset>
  if (strncasecmp(hsdata,"<reset>",7)==0) {
    hsdata+=7;
    while (isspace(*hsdata))
      hsdata++;
    reset=TRUE;
  }

  // hsname is the start of the nfs request
  hsname=hscopy;

  // get either the handle of the old headerset if it exists, or the handle of
  // the hset dir if not existing headerset.
  if (!DirLookup(HsetHandle,hsname,oldhandle))
    HandleCopy(oldhandle,HsetHandle);

  // get the url & headers of the old headerset (or hset dir)
  HandleToUrl(oldhandle,FHSIZE,&url,&inode,&metaptr);
  HandleToHeaders(oldhandle,FHSIZE,metaptr,&oldheaders);

  if (!reset) {
    HsetUpdate(&newheaders,oldheaders,hsdata);
  } else {
    HsetUpdate(&newheaders,NULL,hsdata);
  }

  fprintf(stderr,"%s-%s-%s\n",newheaders,oldheaders,hsdata);
  SafeFree(oldheaders);

  UrlToHandle(url,InodeCreate(ROOTINODE,TRUE),hshandle,FHSIZE,&metaptr);
  HeadersToHandle(newheaders,hshandle,FHSIZE,metaptr);
  HandleSetType(hshandle,FLAG_HEADERSET);

  DirAddChild(HsetHandle,hshandle,hsname,TRUE);
  DirSetParent(hshandle,HsetHandle,hsname);

  SafeFree(url);
  SafeFree(hscopy);

  RecordStatistics("nfs_write",stat);

  result->status = NFS_OK;
  return (result);
}

readres *nfsproc_read_2(readargs *argp, struct svc_req *rqstp)
{
  readres *result = (readres*) AllocResult(sizeof(readres)+sizeof(struct THttpData)+argp->count+32);
  pStatInfo stat;
  pHttpData data=  (pHttpData) (((char*)result)+sizeof(readres));
  char buf[256];
  int success;

  data->ResultPtr=NULL;
  data->Size=argp->count;
  data->Bytes=((char*)result)+sizeof(readres)+sizeof(struct THttpData);
  data->Count=0;

  stat=StartStatistics("nfs_read");

  sprintf(buf,"of=%d, cnt=%d",argp->offset,argp->count);
  debug("nfs_read",buf,(fhandle*) &argp->file);

//  data.ResultPtr = NULL;
//  data.Size = argp->count+1;
//  data.Bytes = (char*) SafeMalloc(argp->count+1);

  if (HandleGetType(argp->file)==FLAG_HEADERS)
    success=GetHeaders(argp->file,argp->offset,argp->count,data,0);
  else if (HandleGetType(argp->file)==FLAG_INFO) {
    success=TRUE;
    HandleGetInfo(argp->file,FHSIZE,data->Bytes,&data->Count);
    data->ResultPtr=data->Bytes;
  } else
    success=GetBytes(argp->file,argp->offset,argp->count,data,0);

  fprintf(stderr,"%d %d\n",success,data->Count);

  if (success) {
    result->status = NFS_OK;
    result->readres_u.readok.data.nfsdata_val = data->ResultPtr;
    result->readres_u.readok.data.nfsdata_len = data->Count;
  } else
    result->status = NFSERR_NOENT;

  AttrLookupConsistant(argp->file,&result->readres_u.readok.attributes);

  debug("nfs_read complete",NULL,NULL);

  RecordStatistics("nfs_read",stat);

  return(result);
}

readdirres *nfsproc_readdir_2(readdirargs *argp, struct svc_req *rqstp)
{
  readdirres *result = (readdirres*) AllocResult(sizeof(readdirres));
  pentry entry=NULL;
  int cookie, outcount;
  pStatInfo stat;

  stat=StartStatistics("nfs_readdir");

  debug("nfs_readdir",NULL,(fhandle*) &argp->dir);

  DirMakeConsist(argp->dir);

  memcpy(&cookie,argp->cookie,4);
  outcount = DirRead(argp->dir,cookie,&(result->readdirres_u.readdirok.entries));

  if (result<0)
    result->status = NFSERR_NOENT;
  else
    result->status = NFS_OK;

  result->readdirres_u.readdirok.eof = TRUE;

  debug("nfs_readdir complete",NULL,NULL);

  RecordStatistics("nfs_readdir",stat);

  return(result);
}

readlinkres *
nfsproc_readlink_2(fhandleargs *argp, struct svc_req *rqstp)
{
  readlinkres *result = (readlinkres*) AllocResult(sizeof(readlinkres));
  pStatInfo stat;
  char *url, *metaptr;
  int inode;

  stat=StartStatistics("nfs_readlink");

  debug("nfs_readlink",NULL,(fhandle*) &argp->data);

  if (HandleGetType(argp->data)!=FLAG_SYMLINK) {
    result->status=NFSERR_PERM;
    RecordStatistics("nfs_readlink_notsymlink",stat);
    return result;
  }

  if (!HandleToUrl(argp->data,FHSIZE,&url,&inode,&metaptr)) {
    result->status=NFSERR_NOENT;
    RecordStatistics("nfs_readlink_noent",stat);
    return result;
  }

  result->status=NFS_OK;
  result->readlinkres_u.data = url; // how is this disposed???

  RecordStatistics("nfs_readlink",stat);

  return(result);
}

diropres *
nfsproc_create_2(createargs *argp, struct svc_req *rqstp)
{
  debug("nfs_create",NULL,NULL);

  return nfsproc_lookup_2(&argp->where,rqstp);
}

diropres *
nfsproc_mkdir_2(createargs *argp, struct svc_req *rqstp)
{
  debug("nfs_mkdir",NULL,NULL);

  return nfsproc_lookup_2(&argp->where,rqstp);
} 

nfsstat *
nfsproc_remove_2(diropargs *argp, struct svc_req *rqstp)
{
  nfsstat *result = (nfsstat*) AllocResult(sizeof(nfsstat));
  pStatInfo stat;

  stat=StartStatistics("nfs_remove");

  debug("nfs_remove",argp->name,&argp->dir);

  RecordStatistics("nfs_remove",stat);

  return result;
}

statfsres *
nfsproc_statfs_2(fhandleargs *argp, struct svc_req *rqstp)
{
	statfsres *result = (statfsres*) AllocResult(sizeof(statfsres));

        debug("nfs_statfs",NULL,NULL);

        result->status = NFS_OK;
        result->statfsres_u.info.tsize = 4096;
        result->statfsres_u.info.bsize = 4096;
        result->statfsres_u.info.blocks = 1024;
        result->statfsres_u.info.bfree = 0;
        result->statfsres_u.info.bavail = 0;

        debug("nfs_statfs complete",NULL,NULL);

	return(result);
}


