#include "stdinc.h"
#include <errno.h>
#include "buffer.h"
#include "html.h"
#include "httpcon.h"
#include "parse.h"
#include "url.h"
#include "safeget.h"

#ifdef HTTPCON_STATS
#include "stats.h"
#endif

#ifdef WINDOWS
#include <dir.h>
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

int HttpconParseLinks = TRUE;
int HttpconAllowShutdown = TRUE;
int HttpconOnlySaveGoodFiles = FALSE;

#ifdef HTTPCON_MUTEX
MUTEX_TYPE MutexSocket = PTHREAD_MUTEX_INITIALIZER;
#endif

void makedir(char *name, int include_fname)
{
  char *start=name, *end;
#ifndef WINDOWS
  struct stat buf;
#endif

  while (*start==SLASH)
    start++;

  while ((end=strchr(start,SLASH))!=NULL) {
    *end = '\0';

//    fprintf(stderr,"mkdir %s\n",name);
#ifdef WINDOWS
    mkdir(name);
#else
    if (stat(name,&buf)!=0)
      mkdir(name,0777);
#endif

    *end = SLASH;
    start=end+1;
  }

#ifdef WINDOWS
  if (include_fname)
    mkdir(name);
#else
  if ((include_fname) && (stat(name,&buf)!=0))
    mkdir(name,0777);
#endif
}

int open_and_makedir(char *name, int mode, int attr)
{
  int descriptor = open(name,mode,attr);

  if ((descriptor==-1) && ((errno==ENOENT) || (errno==ENOTDIR))) {
    makedir(name,FALSE);
    descriptor = open(name,mode,attr);
  }

  return descriptor;
}

void HttpConError(char *s, int d)
{
  if (d>sys_nerr-1)
    printf("%s: unknown error #%d\n",s,d);
  else
    printf("%s: %s\n",s,sys_errlist[d]);
}

int LookupAddress(char *address, struct in_addr *addr)
{
  unsigned long ip_address;

  #ifdef HTTPCON_STATS
  pStatInfo st=StartStatistics("lookupaddress");
  #endif

  if ((ip_address = inet_addr (address)) != INET_ADDR_FAIL) {
    addr->s_addr = ip_address;
  } else {
    struct hostent *hostentry = NULL;

    #ifdef WINDOWS
    if ((hostentry = gethostbyname(address)) == NULL)
      return -1;
    #else
      char buf[2048];
      struct hostent he1;
      int err=0;
      int result;

      result = gethostbyname_r(address,&he1,buf,sizeof(buf),&hostentry,&err);

      if ((hostentry==NULL) || (result!=0) || (err!=0)) {
        #ifdef HTTPCON_STATS
        RecordStatistics("lookupaddress_fail",st);
        #endif
        return -1;
      }
    #endif

    addr->s_addr = *(ADDRTYPE *) hostentry->h_addr;
  }

  #ifdef HTTPCON_STATS
  RecordStatistics("lookupaddress",st);
  #endif

  return 0;
}

SOCKET InitHttpSocket(char *servername, int port)
{
  struct sockaddr_in inaddr;
  SOCKET sock;

#ifdef HTTPCON_MUTEX
  MUTEX_LOCK(MutexSocket);
#endif

  sock = socket(AF_INET, SOCK_STREAM, 0);

  if (sock==INVALID_SOCKET) {
    printf("%s:%d ",servername,port);
    perror("unable to create socket");
//    HttpConError("unable to create socket",servername,port,strerror());
#ifdef HTTPCON_MUTEX
    MUTEX_UNLOCK(MutexSocket);
#endif
    return -1;
  }

  if (LookupAddress(servername,&inaddr.sin_addr)!=0) {
    fprintf(stderr,"%s:%d ",servername,port);
    perror("unable to map address");
//    HttpConError("unable to map address",servername,port,strerror());
    closesocket(sock);
#ifdef HTTPCON_MUTEX
    MUTEX_UNLOCK(MutexSocket);
#endif
    return -1;
  }

  inaddr.sin_family = AF_INET;
  inaddr.sin_port = htons(port);

  if ((connect (sock,(struct sockaddr *)&inaddr,sizeof(struct sockaddr_in))==SOCKET_ERROR)) {
    printf("%s:%d ",servername,port);
    perror("unable to connect");
    closesocket(sock);
#ifdef HTTPCON_MUTEX
    MUTEX_UNLOCK(MutexSocket);
#endif
    return -1;
  }

#ifdef HTTPCON_MUTEX
  MUTEX_UNLOCK(MutexSocket);
#endif

  return sock;
}

void InitHttpHeaders(struct HttpHeaders *h)
{
  h->ResponseCode = 0;
  h->Count = 0;
  h->BufCount = 0;
  h->BufPtr = h->Buf;
  h->ByteCount = 0;
}

void AddHttpHeaders(struct HttpHeaders *h, char *s)
{
  char *colonpos;
  int len = strlen(s);

  if (h==NULL)
    return;

  if (h->Count>=MAX_HTTP_HEADERS)
    return;

  if (h->BufCount+len+2>=HTTP_HEADERS_BYTES)
    return;

  h->Tag[h->Count] = h->BufPtr;
  h->Data[h->Count] = NULL;
  strcpy(h->BufPtr,s);

  colonpos = strchr(h->BufPtr,':');
  if (colonpos!=NULL) {
    *colonpos = '\0';

    colonpos++;
    while ((*colonpos) && (isspace(*colonpos)))
      colonpos++;

    if (*colonpos)
      h->Data[h->Count] = colonpos;
  }

  h->BufPtr+=len+1;
  h->BufCount+=len+1;

  h->Count++;
}

int GetHttpHeaders(struct Buffer *buf, int state, struct HttpHeaders *hdrin)
{
  while (state == HTTP_STATE_HEADERS) {
    if (!BufferGetLine(buf)) {
      state = HTTP_STATE_DONE;
    } else if (*buf->LineStart=='\0') {
      state = HTTP_STATE_DATA;
    } else if (hdrin!=NULL) {
      AddHttpHeaders(hdrin,buf->LineStart);
    }
  }

  return state;
}

/* returns a response code */
int RequestUrlLow(char *proxyname,
               int proxyport,
               char *urlin,
               char *outfn,
               struct tnode **htmlroot,
               char *hdrout,
               struct HttpHeaders *hdrin,
               struct Buffer *buf,
               int timeout)
{
  SOCKET sock;
  char server[MAX_URL_LEN], url[MAX_URL_LEN];
  int port;
  char msg[MAX_URL_LEN];
  int state;
  int did_alloc_buf;
  int respcode = 0;

  if (strlen(urlin)>=MAX_URL_LEN)
    return ERR_REQUEST_TOOLONG;

  if ((proxyname!=NULL) && (*proxyname!='\0')) {
    strcpy(url,urlin);
    strcpy(server,proxyname);
    port=proxyport;
  } else {
    ParseName(urlin,server,&port,url);
  }

  if (htmlroot!=NULL)
    *htmlroot = NULL;

  sock = InitHttpSocket(server,port);

  if (sock==INVALID_SOCKET)
    return ERR_REQUEST_NOSOCKET;

  sprintf(msg,"GET %s HTTP/1.0\r\n",url);
  if (send(sock,msg,strlen(msg),0)==-1)
    return ERR_REQUEST_CANTSENDGET;

  if ((hdrout!=NULL) && (*hdrout))
    if (send(sock,hdrout,strlen(hdrout),0)==-1)
      return ERR_REQUEST_CANTSENDHEADER;

  if (send(sock,"\r\n",2,0)==-1)
    return ERR_REQUEST_CANTSENDHEADER;

  if (HttpconAllowShutdown)
    shutdown(sock,1); // shutdown transmitter

  if (buf==NULL) {
    buf = BufferCreate();
    did_alloc_buf=TRUE;
  } else {
    BufferReset(buf);
    did_alloc_buf=FALSE;
  }

  buf->Timeout = timeout;
  buf->Sock = sock;
  state = HTTP_STATE_HEADERS;

  if (BufferGetLine(buf)) {
    char *after;

    strtok(buf->LineStart," \t\r\n"); /* skip http/1.0 */
    after = strtok(NULL," \t\r\n"); /* get token of response code */
    if ((after==NULL) || ((respcode=atoi(after))==0)) {
      // no response code, so maybe the server is just giving us the data

      respcode = 200;
      BufferToBeginning(buf);
      state = HTTP_STATE_DATA;
    }
  } else
    state = HTTP_STATE_DONE;

  if (hdrin!=NULL)
    hdrin->ResponseCode = respcode;

  if (state==HTTP_STATE_HEADERS)
    state = GetHttpHeaders(buf,state,hdrin);

  if (state==HTTP_STATE_DATA) {
    if ((outfn!=NULL) && ((respcode==200) || (!HttpconOnlySaveGoodFiles))) {
      buf->OutFile = open_and_makedir(outfn,O_BINARY | O_CREAT | O_WRONLY | O_TRUNC,S_IWRITE|S_IREAD);
      if (buf->OutFile<0)
        printf("Unable to open %s\n",outfn);
    }

    BufferDump(buf);

    if (htmlroot!=NULL)
      if (HttpconParseLinks)
        HtmlParseLinks(htmlroot,buf);
      else
        HtmlParse(htmlroot,buf);

    BufferDrainConnection(buf);

    if (buf->OutFile!=-1) {
      close(buf->OutFile);
      buf->OutFile=-1;
    }
  }

  if ((!buf->Abort) && (!buf->HadTimeout))
    while (recv(sock,msg,sizeof(msg),0)>0) ; /* recv any remaining data */

  closesocket(sock); /* close the socket */
  buf->Sock = INVALID_SOCKET;

  if (hdrin!=NULL)
    hdrin->ByteCount = buf->TotalBytes;

  if (did_alloc_buf)
    BufferFree(buf);

  return respcode;
}

int RequestUrl(char *proxyname,
               int proxyport,
               char *urlin,
               char *outfn,
               struct tnode **htmlroot,
               char *hdrout,
               struct HttpHeaders *hdrin,
               struct Buffer *buf,
               int timeout,
               int redirlimit)
{
  char url[MAX_URL_LEN];
  int i,j;
  int respcode;

  if ((redirlimit>0) && (hdrin==NULL)) {
    fprintf(stderr,"httpcon: redirlimit>0 && hdrin==NULL\n");
    exit(1);
  }

  if (strlen(url)>=MAX_URL_LEN) {
    fprintf(stderr,"url len\n");
    return FALSE;
  }

  strcpy(url,urlin);

  /* follow a maximum of two redirects */
  for (i=0; i<redirlimit+1; i++) {
    respcode=RequestUrlLow(proxyname,proxyport,url,outfn,htmlroot,hdrout,hdrin,buf,timeout);

    if (hdrin!=NULL)
      hdrin->ResponseCode = respcode;

    /* -1,-2,etc indicates a fatal error */
    if (respcode<0)
      return FALSE;

    /* 200 is success, 304 is not_modified */
    if ((respcode==200) || (respcode==304))
      return TRUE;

    /* we don't have a redirect, or we have an error, so return FALSE */
    if ((hdrin==NULL) || ((respcode!=301) && (respcode!=302)))
      return FALSE;

    /* Find the header that has location: */
    for (j=0; j<hdrin->Count; j++)
      if ((stricmp(hdrin->Tag[j],"location")==0) && (hdrin->Data[j]!=NULL))
        break;

    /* if no location:, then we can't redirect */
    if (j>=hdrin->Count)
      return FALSE;

    if (i<redirlimit) {
      strcpy(url,hdrin->Data[j]);
      InitHttpHeaders(hdrin);
    }
  }

  /* if we get here, then we exceeded the max redirect count */
  return TRUE;
}

