/************************** Start of HUFF.C *************************
 *
 * This is the Huffman coding module used in Chapter 3.
 * Compile with BITIO.C, ERRHAND.C, and either MAIN-C.C or MAIN-E.C
 */


#include "stdinc.h"
#include "bitio.h"
#include "huff.h"

void HuffmanInitTree(struct THuffParams *p, int treenum, int numcodes)
{
  p->NodesSize= (numcodes*2+2)*sizeof(NODE);
  p->CodesSize= (numcodes+1)*sizeof(CODE);
  p->Trees[treenum].counts= (unsigned long*) calloc(numcodes,sizeof(unsigned long)); // 256
  p->Trees[treenum].nodes = (NODE *) malloc(p->NodesSize);
  p->Trees[treenum].codes = (CODE *) malloc(p->CodesSize);  // 257
  p->Trees[treenum].root_node=0;
}

void HuffmanInit(struct THuffParams *p, int numcodes, int terminator, int bits, int trees)
{
  int i;

//  p->counts = (unsigned long*) calloc(numcodes,sizeof(unsigned long)); // 256
//  p->nodes = (NODE *) calloc(numcodes*2+2,sizeof(NODE)); // 514
//  p->codes = (CODE *) calloc(numcodes+1, sizeof(CODE));  // 257
//  p->root_node=0;

  p->numtrees=trees;
  p->Trees=calloc(p->numtrees,sizeof(struct THuffTree));
  for (i=0; i<p->numtrees; i++)
    HuffmanInitTree(p,i,numcodes);

  p->numcodes=numcodes;
  p->terminator=terminator;
  p->bits=bits;
  p->f13=numcodes*2+1; // 513
  p->END_OF_STREAM=numcodes; // 256

  p->scale=256;
}

void HuffmanConfigure(struct THuffParams *p, char *input)
{
  BIT_FILE *bfin;
  int i;

  bfin = OpenInputBitFile(input,-1);

  count_bytes(bfin,p);

  for (i=0; i<p->numtrees; i++) {
    scale_counts(p,i);
    p->Trees[i].root_node = build_tree(p,i);
    convert_tree_to_code(p,0,0,p->Trees[i].root_node,i);
  }

  CloseInputBitFile(bfin);
}

int HuffmanCompressString(char *input, char *output, struct THuffParams *p)
{
  BIT_FILE *bfin, *bfout;

  bfin = OpenInputBitFile(input,-1);
  bfout = OpenOutputBitFile(output);

  compress_data(bfin,bfout,p);

  CloseInputBitFile(bfin);

  return CloseOutputBitFile(bfout);
}

int HuffmanExpandString(char *input, char *output, struct THuffParams *p, int *bread)
{
  BIT_FILE *bfin, *bfout;

  bfin = OpenInputBitFile(input,-1);
  bfout = OpenOutputBitFile(output);

  expand_data(bfin,bfout,p);

  (*bread)=CloseInputBitFile(bfin);

  return CloseOutputBitFile(bfout);
}

void count_bytes(BIT_FILE *input, struct THuffParams *p)
{
    int c;
    int last = MIN(p->terminator,p->numtrees-1);

    while ((c=InputBits(input,p->bits))!=p->terminator) {
      p->Trees[last].counts[c]++;
      last=MIN(c,p->numtrees-1);
    }
}

void scale_counts(struct THuffParams *p, int treenum)
{
    unsigned long max_count;
    int i;

    max_count = 0;
    for ( i = 0 ; i < p->numcodes ; i++ )
       if ( p->Trees[treenum].counts[ i ] > max_count )
	   max_count = p->Trees[treenum].counts[ i ];

    if ( max_count == 0 ) {
	p->Trees[treenum].counts[ 0 ] = 1;
	max_count = 1;
    }

    max_count = max_count / (p->scale-1);
    max_count = max_count + 1;

    for ( i = 0 ; i < p->numcodes ; i++ ) {
	p->Trees[treenum].nodes[ i ].count = (unsigned int) ( p->Trees[treenum].counts[ i ] / max_count );
	if ( p->Trees[treenum].nodes[ i ].count == 0 /*&& p->trees[treenum].counts[ i ] != 0*/ )
	    p->Trees[treenum].nodes[ i ].count = 1;
    }
    p->Trees[treenum].nodes[ p->END_OF_STREAM ].count = 1;
}

int build_tree(struct THuffParams *p, int treenum)
{
    int next_free;
    int i;
    int min_1;
    int min_2;

    p->Trees[treenum].nodes[p->f13].count = 0xffff;
    for ( next_free = p->END_OF_STREAM + 1 ; ; next_free++ ) {
	min_1 = p->f13;
	min_2 = p->f13;
	for ( i = 0 ; i < next_free ; i++ )
            if ( p->Trees[treenum].nodes[ i ].count != 0 ) {
                if ( p->Trees[treenum].nodes[ i ].count < p->Trees[treenum].nodes[ min_1 ].count ) {
                    min_2 = min_1;
                    min_1 = i;
                } else if ( p->Trees[treenum].nodes[ i ].count < p->Trees[treenum].nodes[ min_2 ].count )
                    min_2 = i;
            }
	if ( min_2 == p->f13 )
	    break;
	p->Trees[treenum].nodes[ next_free ].count = p->Trees[treenum].nodes[ min_1 ].count
	                           + p->Trees[treenum].nodes[ min_2 ].count;
        p->Trees[treenum].nodes[ min_1 ].saved_count = p->Trees[treenum].nodes[ min_1 ].count;
        p->Trees[treenum].nodes[ min_1 ].count = 0;
        p->Trees[treenum].nodes[ min_2 ].saved_count =  p->Trees[treenum].nodes[ min_2 ].count;
        p->Trees[treenum].nodes[ min_2 ].count = 0;
	p->Trees[treenum].nodes[ next_free ].child_0 = min_1;
	p->Trees[treenum].nodes[ next_free ].child_1 = min_2;
    }
    next_free--;
    p->Trees[treenum].nodes[ next_free ].saved_count = p->Trees[treenum].nodes[ next_free ].count;
    return( next_free );
}

void convert_tree_to_code(struct THuffParams *p, unsigned int code_so_far, int bits, int node, int treenum)
{
    if ( node <= p->END_OF_STREAM ) {
	p->Trees[treenum].codes[ node ].code = code_so_far;
	p->Trees[treenum].codes[ node ].code_bits = bits;
	return;
    }
    code_so_far <<= 1;
    bits++;
    convert_tree_to_code( p, code_so_far, bits, p->Trees[treenum].nodes[ node ].child_0,treenum );
    convert_tree_to_code( p, code_so_far | 1,  bits, p->Trees[treenum].nodes[ node ].child_1,treenum );
}

void compress_data(BIT_FILE *input, BIT_FILE *output, struct THuffParams *p)
{
    int c;
    int last = MIN(p->terminator,p->numtrees-1);

    while ((c = InputBits(input,p->bits)) != p->terminator) {
        OutputBits( output, (unsigned long) p->Trees[last].codes[ c ].code,
                    p->Trees[last].codes[ c ].code_bits );
        last = MIN(c,p->numtrees-1);
    }

    OutputBits( output, (unsigned long) p->Trees[last].codes[ p->END_OF_STREAM ].code,
		p->Trees[last].codes[ p->END_OF_STREAM ].code_bits );
}

void expand_data(BIT_FILE *input, BIT_FILE *output, struct THuffParams *p)
{
    int node;
    int last = MIN(p->terminator,p->numtrees-1);

    for ( ; ; ) {
        node = p->Trees[last].root_node;
        do {
            if ( InputBit( input ) )
                node = p->Trees[last].nodes[ node ].child_1;
            else
                node = p->Trees[last].nodes[ node ].child_0;
        } while ( node > p->END_OF_STREAM );

	if ( node == p->END_OF_STREAM )
            break;

        OutputBits(output,node,p->bits);

        last = MIN(node,p->numtrees-1);
    }

    OutputBits(output,p->terminator,p->bits);
}

