#include "stdinc.h"
#include <assert.h>
#include "lru.h"

pLRUInfo LRUInit(int size)
{
  pLRUInfo lru = (pLRUInfo) malloc(sizeof(struct LRUInfo));

  lru->FreeList = NULL;
  lru->Size = size;
  lru->Header = (pLRUNode) malloc(size);

  assert(lru->Header!=NULL);
  lru->Header->Next = lru->Header;
  lru->Header->Prev = lru->Header;

  lru->Count=0;
  lru->MinTime=0;

  return lru;
}

void LRUSetMinTime(pLRUInfo lru, int time)
{
  lru->MinTime = time;
}

int LRUEmpty(pLRUInfo lru)
{
  assert(lru!=NULL);
  assert(lru->Header!=NULL);

  return (lru->Header->Next == lru->Header->Prev);
}

void LRUPromote(pLRUInfo lru, pLRUNode n)
{
  pLRUNode cur;
  int rcount=0;

  assert(lru!=NULL);
  assert(lru->Header!=NULL);
  assert(n!=NULL);
  assert(n!=lru->Header);
  assert(!n->IsOnFreeList);

  #ifdef LRUDEBUG
    printf("promote %d\n",n);
  #endif

  /* unlink n */
  if (n->Next!=NULL) {
    assert(n->Prev!=NULL);
    n->Prev->Next = n->Next;
    n->Next->Prev = n->Prev;
  } else
    lru->Count++;

  #ifdef SLOW_DEBUG_CODE
  for (cur=lru->Header->Next; cur!=lru->Header; cur=cur->Next) {
    assert(cur!=n);
    rcount++;
    if (rcount>lru->Count*10) {
      fprintf(stderr,"LRU: rcount=%d, lru->Count = %d\n",rcount,lru->Count);
      assert(FALSE);
    }
  }
  #endif

  /* relink n at the front of the list */
  n->Next = lru->Header->Next;
  n->Prev = lru->Header;
  lru->Header->Next->Prev = n;
  lru->Header->Next = n;

  if (lru->MinTime>0)
    SET_TIMESTAMP(n->PromoteTime);
}

int LRUExists(pLRUNode n)
{
  assert(n!=NULL);

  if (n->Next!=NULL) {
    assert(n->Prev!=NULL);
    return TRUE;
  } else {
    assert(n->Prev==NULL);
    return FALSE;
  }
}

void LRUDelete(pLRUInfo lru, pLRUNode n)
{
  pLRUNode cur;
  int rcount=0;

  assert(n!=NULL);
  assert(n!=lru->Header);

  assert(!n->IsOnFreeList);
  assert(n->Next!=n); /* make sure it's not the header */
  assert(n->Prev!=n);
  assert(lru->Count>0);

  #ifdef LRUDEBUG
    printf("delete %d\n",n);
  #endif

  n->Next->Prev = n->Prev;
  n->Prev->Next = n->Next;

  n->Next = NULL;
  n->Prev = NULL;

  lru->Count--;

  #ifdef SLOW_DEBUG_CODE
  for (cur=lru->Header->Next; cur!=lru->Header; cur=cur->Next) {
    assert(cur!=n);
    rcount++;
    if (rcount>lru->Count*10) {
      fprintf(stderr,"LRU: rcount=%d, lru->Count = %d\n",rcount,lru->Count);
      assert(FALSE);
    }
  }
  #endif
}

pLRUNode LRULeastRecent(pLRUInfo lru)
{
  assert(lru);

  if (LRUEmpty(lru))
    return NULL;

  assert(!lru->Header->Prev->IsOnFreeList);

  return lru->Header->Prev;
}

pLRUNode LRUMostRecent(pLRUInfo lru)
{
  assert(lru);

  if (LRUEmpty(lru))
    return NULL;

  assert(!lru->Header->Next->IsOnFreeList);

  return lru->Header->Next;
}

pLRUNode LRUPopLeastRecent(pLRUInfo lru)
{
  pLRUNode n = LRULeastRecent(lru);

  if (n==NULL)
    return NULL;

  if (lru->MinTime>0) {
    TIMESTAMP finish;

    SET_TIMESTAMP(finish);
    if (GET_ELAPSED(n->PromoteTime,finish)<lru->MinTime)
      return NULL;
  }

  LRUDelete(lru,n);

  return n;
}

pLRUNode LRUAllocNode(pLRUInfo lru)
{
  pLRUNode result;

  if (lru->FreeList!=NULL) {
    result = lru->FreeList;
    lru->FreeList = lru->FreeList->Next;
    assert(result->IsOnFreeList);
  } else
    result = (pLRUNode) malloc(lru->Size);

  result->Next = NULL;
  result->Prev = NULL;
  result->Data = NULL;
  result->IsOnFreeList = FALSE;

  assert(!result->IsOnFreeList);

  #ifdef LRUDEBUG
    printf("alloc %d\n",result);
  #endif

  return result;
}

void LRUFreeNode(pLRUInfo lru, pLRUNode n)
{
  #ifdef LRUDEBUG
    printf("free %d\n",n);
  #endif

  n->Next = lru->FreeList;
  lru->FreeList=n;

  n->IsOnFreeList = TRUE;
}

void LRUDestroy(pLRUInfo lru)
{
  pLRUNode n;

  assert(lru!=NULL);
  assert(lru->Header!=NULL);

  while ((n=LRUPopLeastRecent(lru))!=NULL)
    free(n);

  while (lru->FreeList!=NULL) {
    n=lru->FreeList->Next;
    free(n);
    lru->FreeList=n;
  }

  free(lru->Header);
  free(lru);
}

int LRUGetCount(pLRUInfo lru)
{
  pLRUNode cur;
  int rcount = 0;

  #ifdef SLOW_DEBUG_CODE
  for (cur=lru->Header->Next; cur!=lru->Header; cur=cur->Next) {
    assert(cur!=NULL);
    rcount++;

    if (rcount>lru->Count*10) {
      fprintf(stderr,"LRU: rcount=%d, lru->Count = %d\n",rcount,lru->Count);
      assert(FALSE);
    }
  }

  if (rcount!=lru->Count) {
    fprintf(stderr,"LRU: rcount=%d, lru->Count = %d\n",rcount,lru->Count);
    assert(FALSE);
  }
  #endif

  return lru->Count;
}



