/* LRU Package, Scott M Baker, 1998

   Implements LRU using a linked list.

   LRUInit(size) creates a LRU structure with nodes of the given size. Size
     should be == sizeof(struct LRUNode) for the default entries or can be
     increased if additional static data is desired in the node structure.

   LRUPromote(lru,node) will remove node from it's current position and move
     it to the top of the list. If node is not currently in the list then it is
     only inserted at the top, not removed.

   LRUPopLeastRecent(lru) will remove and return the least recent (bottom) node
     in the list, or NULL if the list is empty.

   LRUAllocNode(lru) and LRUFreeNode(lru,node) implement a freelist based
     allocation.

   LRUDestroy(lru) destroys the whole thing and frees all memory.

   LRUSetMinTime(lru) sets a minimum time which must elapsed after a node
     is promoted (LRUPromote) before it is popped (LRUPopLeastRecent). If
     this time has not elapsed, then LRUPopLeastRecent() will return NULL.
*/


/* make sure include only once */
#ifndef _LRU_H
#define _LRU_H

#ifdef __CPLUSPLUS__
extern "C" {
#endif

/* the list node structure typedef */
typedef struct LRUNode {
    struct LRUNode *Next;
    struct LRUNode *Prev;
    void *Data;
    T_TIMESTAMP PromoteTime;
    int IsOnFreeList;
} LRUNode, *pLRUNode;

typedef struct LRUInfo {
  struct LRUNode *Header;
  struct LRUNode *FreeList;
  int Size;
  int Count;
  int MinTime;
} LRUInfo, *pLRUInfo;

#define LRU_NODE_BASE_SIZE (sizeof(struct LRUNode))

/* ------------------------------ Prototypes ------------------------------ */
pLRUInfo LRUInit(int size);
int LRUEmpty(pLRUInfo lru);
void LRUPromote(pLRUInfo lru, pLRUNode n);
void LRUDelete(pLRUInfo lru, pLRUNode n);
pLRUNode LRULeastRecent(pLRUInfo lru);
pLRUNode LRUMostRecent(pLRUInfo lru);
pLRUNode LRUPopLeastRecent(pLRUInfo lru);
pLRUNode LRUAllocNode(pLRUInfo lru);
void LRUFreeNode(pLRUInfo lru, pLRUNode n);
void LRUDestroy(pLRUInfo lru);
int LRUGetCount(pLRUInfo lru);
int LRUExists(pLRUNode n);
void LRUSetMinTime(pLRUInfo lru, int time);

#ifdef __CPLUSPLUS__
}
#endif

/* ----------------------------- End of list.h --------------------------- */
#endif /* _LRU_H */


