/************************** Start of LZW12.C *************************
 *
 * This is 12 bit LZW program, which is discussed in the first part
 * of the chapter.  It uses a fixed size code, and does not attempt
 * to flush the dictionary after it fills up.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include "errhand.h"
#include "bitio.h"
#include "lzw12.h"

#ifndef FALSE
#define FALSE (1==0)
#define TRUE (!FALSE)
#endif

/*
 * Constants used throughout the program.  BITS defines how many bits
 * will be in a code.  TABLE_SIZE defines the size of the dictionary
 * table.
 */
#define BITS                       13
#define MAX_CODE                   ( ( 1 << BITS ) - 1 )
#define END_OF_STREAM              256
#define FIRST_CODE                 257
#define UNUSED                     -1

/*
 * Local prototypes.
 */

unsigned int find_child_node( int parent_code, int child_character, struct TLZWParams *p);
unsigned int decode_string( unsigned int offset, unsigned int code, struct TLZWParams *p);

/*
 * This data structure defines the dictionary.  Each entry in the dictionary
 * has a code value.  This is the code emitted by the compressor.  Each
 * code is actually made up of two pieces:  a parent_code, and a
 * character.  Code values of less than 256 are actually plain
 * text codes.
 */


//char decode_stack[ TABLE_SIZE ];

/*
 * The compressor is short and simple.  It reads in new symbols one
 * at a time from the input file.  It then  checks to see if the
 * combination of the current symbol and the current code are already
 * defined in the dictionary.  If they are not, they are added to the
 * dictionary, and we start over with a new one symbol code.  If they
 * are, the code for the combination of the code and character becomes
 * our new code.
 */

void LZWCompressInit(struct TLZWParams *p)
{
  unsigned int i;

  p->next_code = FIRST_CODE;

  for ( i = 0 ; i < TABLE_SIZE ; i++ )
    p->dict[i].code_value = UNUSED;
}

int LZWCompressString(char *input, struct TLZWParams *p, char *dest)
{
  BIT_FILE *bf;

  bf = OpenOutputBitFile(dest);
  LZWCompressFile(input,strlen(input),bf,p);
  return CloseOutputBitFile(bf);
}

void LZWExpandString(char *input, int count, struct TLZWParams *p, char *dest, int *bread)
{
  BIT_FILE *bf;
  int bwritten;

  bf = OpenInputBitFile(input,count);
  bwritten = LZWExpandFile(bf,dest,p);
  *bread = CloseInputBitFile(bf);

  *(dest+bwritten)='\0';
}

void LZWCompressFile(char *input, int bytes, BIT_FILE *output, struct TLZWParams *p)
{
    int character;
    int string_code;
    unsigned int index;

    if (bytes<=0) // (string_code = *(input++)) == '\0' )
      string_code = END_OF_STREAM;
    else
      string_code = (unsigned char) *(input++);
      bytes--;
      while ((bytes--)>0) { // (character = *(input++)) != '\0' ) {
        character = (unsigned char) *(input++);

        index = find_child_node( string_code, character, p);
        if ( p->dict[ index ].code_value != -1 )
            string_code = p->dict[ index ].code_value;
        else {
            if ( p->next_code <= MAX_CODE ) {
                p->dict[ index ].code_value = p->next_code++;
                p->dict[ index ].parent_code = string_code;
                p->dict[ index ].character = (char) character;
            }
            OutputBits( output, (unsigned long) string_code, BITS );
            string_code = character;
        }
      }
    OutputBits( output, (unsigned long) string_code, BITS );
    OutputBits( output, (unsigned long) END_OF_STREAM, BITS );
}

void LZWExpandInit(struct TLZWParams *p)
{
  p->did_old_code = FALSE;
  p->next_code = FIRST_CODE;
}

int LZWExpandFile(BIT_FILE *input, char *output, struct TLZWParams *p)
{
    unsigned int new_code;
    unsigned int count;
    char *origoutput = output;

    if (p->did_old_code==FALSE) {
      p->did_old_code=TRUE;

      p->old_code = (unsigned int) InputBits( input, BITS );
      if ( p->old_code == END_OF_STREAM )
          return (output-origoutput);

      p->character = p->old_code;

      *(output++)=(char)p->old_code;
    }

    while ( ( new_code = (unsigned int) InputBits( input, BITS ) )
              != END_OF_STREAM ) {
/*
** This code checks for the CHARACTER+STRING+CHARACTER+STRING+CHARACTER
** case which generates an undefined code.  It handles it by decoding
** the last code, and adding a single character to the end of the decode string.
*/
        if ( new_code >= p->next_code ) {
            p->decode_stack[ 0 ] = (char) p->character;
            count = decode_string( 1, p->old_code, p);
        } else
            count = decode_string( 0, new_code, p);

        p->character = p->decode_stack[ count - 1 ];
        while ( count > 0 ) 
            *(output++) = p->decode_stack[--count];

        if ( p->next_code <= MAX_CODE ) {
            p->dict[ p->next_code ].parent_code = p->old_code;
            p->dict[ p->next_code ].character = (char) p->character;
            p->next_code++;
        }
        p->old_code = new_code;
    }

//    *(output)='\0';

    return (output-origoutput);
}

/*
 * This hashing routine is responsible for finding the table location
 * for a string/character combination.  The table index is created
 * by using an exclusive OR combination of the prefix and character.
 * This code also has to check for collisions, and handles them by
 * jumping around in the table.
 */
unsigned int find_child_node(int parent_code, int child_character, struct TLZWParams *p)
{
    int index;
    int offset;

    index = ( child_character << ( BITS - 8 ) ) ^ parent_code;
    if ( index == 0 )
        offset = 1;
    else
        offset = TABLE_SIZE - index;
    for ( ; ; ) {
        if ( p->dict[ index ].code_value == UNUSED )
            return( index );
        if ( p->dict[ index ].parent_code == parent_code &&
             p->dict[ index ].character == (char) child_character )
            return( index );
        index -= offset;
        if ( index < 0 )
            index += TABLE_SIZE;
    }
}

/*
 * This routine decodes a string from the dictionary, and stores it
 * in the decode_stack data structure.  It returns a count to the
 * calling program of how many characters were placed in the stack.
 */

unsigned int decode_string(unsigned int count, unsigned int code, struct TLZWParams *p)
{
    while ( code > 255 ) {
        p->decode_stack[ count++ ] = p->dict[ code ].character;
        code = p->dict[ code ].parent_code;
    }
    p->decode_stack[ count++ ] = (char) code;
    return( count );
}

/************************** End of LZW12.C *************************/

