#ifdef WINDOWS
#include <windows.h>
#endif

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "smalloc.h"

#ifdef SAFE_MALLOC_MUTEX
#include "thread.h"
#endif

#define MAX_MALLOC_INFO 256

#define MAGIC 45672345L

#ifndef TRUE
#define TRUE 1
#define FALSE (!TRUE)
#endif

typedef struct TMallocInfo {        
  char *Filename;
  int Line;
  unsigned long int Malloc, Free;
  unsigned long int MallocCount, FreeCount;
} TMallocInfo;

TMallocInfo MallocInfo[MAX_MALLOC_INFO];
int MallocInfoCount=0;
int Initialized=0;

#ifdef SAFE_MALLOC_MUTEX
MUTEX_TYPE MutexMalloc;
#endif

void SafeMallocLock(void)
{
  #ifdef SAFE_MALLOC_MUTEX

  MUTEX_LOCK(MutexMalloc);
  #endif
}

void SafeMallocUnlock(void)
{
  #ifdef SAFE_MALLOC_MUTEX
  assert(Initialized);
  MUTEX_UNLOCK(MutexMalloc);
  #endif
}


void SafeMallocDump(void)
{
  int i;
  FILE *fout = fopen("smalloc.txt","wt");

  if (fout==NULL)
    return;

  SafeMallocLock();

  fprintf(fout,"module              line  alloc      free       out        #alloc     #free\n");
  for (i=0; i<MallocInfoCount; i++) {
    char *name=MallocInfo[i].Filename;

    if ((name!=NULL) && (strrchr(name,'\\')!=NULL))
      name = strrchr(name,'\\')+1;

    fprintf(fout,"%-18s %5d %10d %10d %10d %10d %10d\n",
      name,MallocInfo[i].Line,MallocInfo[i].Malloc,
      MallocInfo[i].Free,MallocInfo[i].Malloc-MallocInfo[i].Free,
      MallocInfo[i].MallocCount,MallocInfo[i].FreeCount);
  }

  fclose(fout);

  SafeMallocUnlock();
}

void SafeMallocInit()
{
  #ifdef SAFE_MALLOC_MUTEX
  CREATE_MUTEX(MutexMalloc);
  #endif
  Initialized=TRUE;
}


int AddMallocInfo(int alloc, char *fn, int line)
{
  int i;

  for (i=0; i<MallocInfoCount; i++)
    if ((MallocInfo[i].Filename==fn) && (MallocInfo[i].Line==line))
      break;

  if (i>=MAX_MALLOC_INFO)
    return -1;

  if (i==MallocInfoCount) {
    MallocInfo[i].Filename=fn;
    MallocInfo[i].Line=line;
    MallocInfo[i].Malloc=0;
    MallocInfoCount++;
  }

  MallocInfo[i].Malloc+=alloc;
  MallocInfo[i].MallocCount++;

  return i;
}

void SetFreeInfo(void *x)
{
  int *base = (int*)x;
  int magic = *(base-3);
  int index = *(base-2);//(int*)(x+4);
  int size = *(base-1);//*(((int*)(x))+1);

  if (magic!=MAGIC)
    assert(magic==MAGIC);

  if (index==-1)
    return;

  assert(index>=0);
  assert(index<MAX_MALLOC_INFO);

  MallocInfo[index].Free+=size;
  MallocInfo[index].FreeCount++;
}

void *SafeMalloc(int size)
{
  return SafeMalloc1(size,NULL,0);
}

void *SafeMalloc1(int size, char *fn, int line)
{
  void *result;
#ifdef SAFE_MALLOC_TRACE
  int *base;
  int sizetoalloc = size+12;
#else
  int sizetoalloc = size;
#endif

  SafeMallocLock();

#ifdef WINDOWS
  result = HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,sizetoalloc);
#else
  result = malloc(sizetoalloc);
#endif

  assert(result!=NULL);

#ifdef SAFE_MALLOC_TRACE
  base=(int*)result;
  *(base) = MAGIC;
  *(base+1) = AddMallocInfo(size,fn,line);
  *(base+2) = size;
  result=(void*) (((char*)result)+12);
#endif

  SafeMallocUnlock();

  return result;
}

void SafeFree(void *x)
{
#ifdef SAFE_MALLOC_TRACE
  void *ptrtofree = (void*) (((char*)x)-12);
#else
  void *ptrtofree = x;
#endif

  SafeMallocLock();

#ifdef SAFE_MALLOC_TRACE
  SetFreeInfo(x);
#endif

  #ifdef WINDOWS
  assert(HeapFree(GetProcessHeap(),0,ptrtofree));
  #else
  free(ptrtofree);
  #endif

  SafeMallocUnlock();
}

char *SafeStrDup1(char *x, char *file, int line)
{
  char *dest;

  assert(x!=NULL);

  dest=SafeMalloc1(strlen(x)+1,file,line);
  strcpy(dest,x);

  return dest;
}

char *SafeStrDup(char *x)
{
  return SafeStrDup1(x,NULL,0);
}




